# coding=utf-8
from __future__ import absolute_import, division, print_function

__metaclass__ = type

import os.path as path
from astwro.tools.gapick import main
from astwro.utils.TmpDir import TmpDir
from astwro.starlist import read_dao_file, read_ds9_regions

def test_gapick_short():
    d = TmpDir()
    r = main(
        ga_max_iter = 2, # just two iteration of GA
        ga_pop = 15,  # small population
        overwrite = True,
        out_dir = d.path,
        fine = False,
        demo=True,
    )
    assert (r.stars_number() > 10)

    ap = read_dao_file(path.join(d.path, 'i.ap'))
    lst = read_dao_file(path.join(d.path, 'gen_last.lst'))
    coo = read_dao_file(path.join(d.path, 'i.coo'))
    nei = read_dao_file(path.join(d.path, 'i.nei'))
    reg = read_ds9_regions(path.join(d.path, 'gen_last.reg'))




# coding=utf-8
from collections import namedtuple
from itertools import chain

class DAO(object):
    FType = namedtuple('DAOFileType', ['columns', 'extension', 'NL', 'read_cols'])
    class CType:
        """Column specification"""
        def __init__(self, name, format, NaN, default=None, optional=False):
            self.name = name
            self.format = format
            self.NaN = NaN
            self.default = default
            self.optional = optional

    UNKNOWN_FILE = FType(
        columns   = ['id'] + list(range(1, 100)),  # id then unknown columns
        extension = '.stars',
        NL = 1,
        read_cols=None,
    )
    COO_FILE = FType(
        columns   = ['id', 'x', 'y', 'rmag', 'fsharp', 'round', 'mround'],
        extension = '.coo',
        NL = 1,
        read_cols=None,
    )
    AP_FILE_ODD = FType(   # odd rows columns (used in read)
        columns   = ['id', 'x', 'y', 'mag'] + ['A{:X}'.format(n) for n in range(2,13)],  # id,x,y,mag,A2,A3,..,AC
        extension = '',
        NL = 2,
        read_cols=None,
    )
    AP_FILE_EVEN = FType(  # even rows columns (used in read)
        columns   = ['sky', 'sky_err', 'sky_skew', 'mag_err'] + ['A{:X}_err'.format(n) for n in range(2,13)],
        extension='',
        NL=2,
        read_cols=None,
    )
    AP_FILE = FType(  # (used in write)
        # id,x,y,mag,A2,..,AC,asky,asky_err,asky_skew,mar_err, A2_err,A3_err,...,AC_err
        columns   = AP_FILE_ODD.columns + AP_FILE_EVEN.columns,
        extension = '.ap',
        NL = 2,
        read_cols=None,
    )
    LST_FILE = FType(
        columns   = ['id', 'x', 'y', 'mag', 'mag_err', 'd'],
        extension = '.lst',
        NL = 3,
        read_cols=None,
    )
    NEI_FILE = FType(
        columns   = ['id', 'x', 'y', 'mag', 'sky'],
        extension = '.nei',
        NL = 3,
        read_cols=5,
    )
    ALS_FILE = FType(
        columns   = ['id', 'x', 'y', 'mag', 'mag_err', 'sky', 'iter', 'chi', 'sharp'],
        extension = '.als',
        NL = 1,
        read_cols=None,
    )
    ERR_FILE = FType(
        columns   = ['id', 'err'],
        extension = '.err',
        NL=None,
        read_cols = 2,
    )
    SHORT_FILE = FType(   # Filetype for all_stars files in Bialkow workflow
        columns=['id', 'x', 'y', 'mag', 'mag_err'],
        extension='.all_stars',
        NL=1,
        read_cols=None,
    )
    XY_FILE = FType(   # Filetype for position x,y only files
        columns=['id', 'x', 'y'],
        extension='.xy',
        NL=1,
        read_cols=3,
    )
    SKY_FILE = FType(   # Filetype for WCS coordinates file as used by ccd_phot workflow
        columns=['id', 'x', 'y', 'mag', 'mag_err', 'ra', 'dec'],
        extension='.sky',
        NL=1,
        read_cols=None,
    )
    RADEC_FILE = FType(   # Filetype for WCS coordinates
        columns=['id', 'ra', 'dec'],
        extension='.radec',
        NL=1,
        read_cols=None,
    )


    file_types = {
        COO_FILE.extension: COO_FILE,
        AP_FILE.extension: AP_FILE,
        LST_FILE.extension: LST_FILE,
        NEI_FILE.extension: NEI_FILE,
        ALS_FILE.extension: ALS_FILE,
        ERR_FILE.extension: ERR_FILE,
        SHORT_FILE.extension: SHORT_FILE,
        XY_FILE.extension: XY_FILE,
        RADEC_FILE.extension: RADEC_FILE,
    }


    # columns dictionary, the key is either:
    # - tuple (filetype:FType, column:str)
    # - column:str
    # when searching for column details, first lookup is for pair (filetype:FType.extension, column:str), if not found
    # lookup for column:str (column definition shared by file types), eventually for '_default'

    _static_columns = {
        '_default':     CType('_default', ' {:8.3f}', [-9.999]),
        'id':           CType('id', ' {:6.0f}', [0]),
        ('.als','mag'): CType('mag', ' {:8.4f}', [99.999]),
        ('.als','mag_err'):   CType('mag_err', ' {:8.4f}', [9.9999], default=0.0),
        ('.sky','mag'): CType('mag', '{:8.4f}', [99.999]),
        ('.sky','mag_err'):   CType('mag_err', ' {:8.4f}', [9.9999], default=0.0),
        'iter':         CType('iter', ' {:7.0f}.', [-9.999], default=0.0),
        'chi':          CType('chi', ' {:8.3f}', [-9.999], default=0.0),
        'sharp':        CType('sharp', ' {:8.3f}', [-9.999], default=0.0),
        'sky':          CType('sky', ' {:9.3f}', [-9.999]),
        'sky_err':      CType('sky_err', ' {:5.2f}', [-9.99, -9999.0], default=0),
        'sky_skew':     CType('sky_skew', ' {:5.2f}', [-9.99], default=0),
        'ra':           CType('ra', ' {:s}', [], default='00:00:00.0000'),
        'dec':          CType('dec', ' {:s}', [], default='+00:00:00.000'),
        ('.ap','mag'):  CType('mag', ' {:8.3f}', [99.999,-99.999,94.999, 98.999]),
        ('.ap','mag_err'):   CType('mag_err', ' {:8.4f}', [9.9999], default=0.0),
        ('.ap', 'id'):  CType('id', '\n{:7.0f}', [0]),         # new linie in AP files
        ('.ap', 'sky'): CType('sky', '\n{:14.3f}', [-9.999,0]),  # new linie in AP files
    }

# add A2, A3,... and A2_err, A3_err,...
DAO._apert_columns = dict([(col, DAO.CType(col, ' {:8.3f}', [99.999,-99.999,94.999], optional=True))
                       for col in ('A{:X}'.format(i) for i in range(2,13))])
DAO._ap_err_columns = dict([(col, DAO.CType(col, ' {:7.4f}', [9.9999], optional=True))
                        for col in ('A{:X}_err'.format(i) for i in range(2,13))])
# dict union
DAO.columns = dict(chain.from_iterable(d.items() for d in (DAO._static_columns, DAO._apert_columns, DAO._ap_err_columns)))




import pandas as pd
import numpy as np
from astropy.table import Table
from astropy.coordinates import SkyCoord
import astropy.units as u
from .fileformats import DAO

# Subclassing pandas:
# http://pandas.pydata.org/pandas-docs/stable/internals.html#subclassing-pandas-data-structures

class StarList(pd.DataFrame):
    #
    # StarList properties extending  pd.DataFrame
    _metadata = ['_DAO_hdr', '_DAO_type']
    _DAO_hdr = None
    _DAO_type = None

    @staticmethod
    def new():
        """Returns empty StarList instance with columns id,x,y"""
        idx = pd.Series(name='id', dtype='int64')
        id = pd.Series(dtype='int64')
        x = pd.Series(dtype='float64')
        y = pd.Series(dtype='float64')
        return StarList({'id': id, 'x': x, 'y': y}, index=idx)

    @property
    def _constructor(self):

        return StarList

    @property
    def DAO_hdr(self):
        """DAO file header dict if any"""
        return self._DAO_hdr

    @DAO_hdr.setter
    def DAO_hdr(self, hdr):
        self._DAO_hdr = hdr

    @property
    def DAO_type(self):
        """DAO file header dict if any"""
        return self._DAO_type

    @DAO_type.setter
    def DAO_type(self, typ):
        self._DAO_type = typ

    def import_metadata(self, src):
        """Copies metdata (dao type, dao hdr) from src

        :param StarList src: source of metadata
        """
        self.DAO_type = src.DAO_type
        self.DAO_hdr = src.DAO_hdr

    def stars_number(self):
        """returns number of stars in list"""
        return self.shape[0]

    def renumber(self, start=1):
        """Renumbers starlist (in place), updating `id` column and index to range start.. start+stars_number"""
        self['id'] = range(start, self.stars_number() + start)
        self.index = self.id

    def refresh_id(self):
        """Cheks for id column existence and renumber if needed then recreate index basin on id

            If `id` column exist but cannot be casted to `int` (by `pandas.Series.to_numeric`),
            `ValueError` exception is raised
        """
        try:
            if self['id'].dtype != int:
                self['id'] = pd.to_numeric(self['id'], downcast='integer', errors='raise')
        except KeyError:
            self.renumber()
        self.index = self['id']


    def to_table(self):
        """
        Return a :class:`astropy.table.Table` instance
        """
        t = Table.from_pandas(self)
        if self.DAO_hdr is not None:
            t.meta.update(self.DAO_hdr)
        if self.DAO_type is not None:
            t.meta['DAO_type'] = self.DAO_type
        return t

    @classmethod
    def from_table(cls, table):
        """
        Create a `StarList` from a :class:`astropy.table.Table` instance

        Parameters
        ----------
        table : :class:`astropy.table.Table`
            The astropy :class:`astropy.table.Table` instance
        Returns
        -------
        sl : `StarList`
            A `StarList`instance
        """
        from .file_helpers import as_starlist
        return as_starlist(table)

    @classmethod
    def from_skycoord(cls, coo):
        """
        Create a `StarList` from a :class:`astropy.coordinates.SkyCoord` instance

        Parameters
        ----------
        coo : :class:`astropy.coordinates.SkyCoord`
            Source
        Returns
        -------
        sl : `StarList`
            A `StarList`instance
        """
        sl = StarList(
            np.array([coo.ra.to_string(sep=':', unit=u.hourangle, pad=True),
                      coo.dec.to_string(sep=':', unit=u.deg, pad=True, alwayssign=True)]).T,
                      columns=['ra', 'dec'])
        sl.DAO_type = DAO.RADEC_FILE
        sl.renumber()
        return sl

    def radec_hmsdms_from_skycoord(self, coo):
        """Uses external `xy2sky` tool """
        self['ra'] =  coo.ra.to_string(sep=':', unit=u.hourangle, pad=True)
        self['dec'] = coo.dec.to_string(sep=':', unit=u.deg, pad=True, alwayssign=True)

    def radec_deg_from_hmsdms(self):
        """Uses external `xy2sky` tool """
        sky = SkyCoord(self.ra, self.dec, unit=(u.hourangle, u.deg))
        self['ra_deg'] = sky.ra.deg
        self['dec_deg'] = sky.dec.deg

    def radec_hmsdms_from_deg(self):
        """Uses external `xy2sky` tool """
        sky = SkyCoord(self.ra_deg, self.dec_deg, unit=u.deg)
        self['ra'] = sky.ra.to_string(sep=':', pad=True, unit=u.hourangle)
        self['dec'] = sky.dec.to_string(sep=':',pad=True, alwayssign=True)

    def to_skycoords(self):
        try:
            sky = SkyCoord(self.ra, self.dec, unit=(u.hourangle, u.deg))
        except:
            sky = SkyCoord(self.ra_deg, self.dec_deg, unit=u.deg)
        return sky

    def xy_from_radec(self, fitsimage):
        from astwro.coord import skyradec2xy
        try:
            x,y = skyradec2xy(self.ra_deg, self.dec_deg, unit='deg', transformer=fitsimage)
        except:
            x,y = skyradec2xy(self.ra, self.dec, unit=(u.hourangle, u.deg),  transformer=fitsimage)
        self['x'] = x
        self['y'] = y

    def magnitudes(self):
        """Returns magnitudes array from columns mag, A2, A3, ..."""
        columns = ['mag'] + list(DAO._apert_columns)
        return self.array_from_columns(columns)


    def magnitudes_err(self):
        """Returns magnitudes errors array from columns mag_err, A2_err, A3_err, ..."""
        columns = ['mag_err'] + list(DAO._ap_err_columns)
        return self.array_from_columns(columns)

    def array_from_columns(self, columns):
        # only existing columns
        existing = self.columns.values.tolist()
        columns = [c for c in columns if c in existing]
        return self[columns].to_numpy()



# coding=utf-8
from __future__ import absolute_import, division, print_function

__metaclass__ = type

from astwro.phot import *
from os.path import join
from astwro.phot.lc_io import write_lc, write_lc_filters
from astwro.utils import TmpDir
import numpy as np
import sys


def prepare_testset():
    s = np.ma.masked_array([12.0, 12.6, 13.0, 13.2, 14.1]).reshape(5, 1)  # 5 stars magnitudes
    o = np.ma.masked_array([1.0, 1.08, 0.9, 1.3]).reshape(1, 4)  # 4 observations deviations
    x = s.dot(o) + np.random.normal(scale=0.1, size=[5, 4])  # 5x4 noisy observations simulated
    return s, o, x


def prepare_testset_masked():
    X = 0.0
    d = np.array([
        [ 11.5,   X,   12.5,   X  ],
        [ 10.0,   X,   11.0,  9.0 ],  # comp
        [ 11.0,   X,     X , 10.0 ],  # comp
        [   X ,   X,     X ,   X  ],
        [   X ,   X,     X ,   X  ],  # comp (bad)
        [ 12.0,   X,   13.0, 11.0 ],
    ])
    e = d * 0.01
    d = np.ma.masked_equal(d, X)
    e = np.ma.masked_equal(e, X)
    d += np.random.normal(scale=0.1, size=d.shape)

    c = [False, True, True, False, True, False]
    return d, e, c

def prepare_testset_real():
    import os
    import pickle
    pickelfile = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'some_stars_data.pkl')
    with open(pickelfile, 'rb') as f:
        if sys.version_info >= (3, 0):
            d, e, c = pickle.load(f, encoding='bytes')
        else:
            d, e, c = pickle.load(f)
    return d, e, c

def test_masked():
    d, e, c = prepare_testset_masked()
    o, s, l, oe, se, le = dphot(d, e, c)
    pass


def test_real():
    d, e, f = prepare_testset_real()
    o, s, l, oe, se, le = dphot_filters(d, e, f)
    pass

def test_masked_object():
    d, e, c = prepare_testset_masked()
    s, l, o, se, le, oe = dphot(d, e, c)
    dp = DiffPhot(d, e, c)
    assert  np.allclose(dp.mag, s)
    assert  np.allclose(dp.lc, l)
    assert  np.allclose(dp.obs_deltas, o)


    e = dp.err
    e2= e**2
    d = dp.obs_deltas_stddev
    d2 = d**2

    x =  (dp.mag, dp.mag_v2, dp.lc_e)
#    assert  np.allclose(dp.mag_e, se)



def test_write_lc():
    #TODO: Limit numer of stars in some_stars_data.pkl
    with TmpDir() as td:
        d, e, f = prepare_testset_real()
        hjd = np.linspace(989.34316, 1002.63264, d.shape[1])
        write_lc(hjd, d, e, prefix=join(td.path, 'lc_'))

def test_write_lc_filter():
    d, e, f = prepare_testset_real()
    hjd = np.linspace(989.34316, 1002.63264, d.shape[1])
    fn = ['B', 'Han']
    r = np.random.uniform(size=d.shape[1])
    fm = np.array([r<0.5, r>=0.5], dtype=bool)
    with TmpDir() as td:
        write_lc_filters(fm, fn, hjd, d, e, prefix=join(td.path, 'lc_'))
        pass



# coding=utf-8
from __future__ import absolute_import, division, print_function
__metaclass__ = type

import os.path as path
from io import StringIO
from astropy.table import Table
import astwro.starlist as sl
import astwro.sampledata as data
from astwro.utils import tmpdir

def test_read_VOT_via_astropy():
    file = path.join(path.abspath(path.dirname(__file__)), 'NGC7142.vot')
    t_astropy = Table.read(file)
    t_starlist = sl.StarList.from_table(t_astropy)







# coding=utf-8
from __future__ import absolute_import, division, print_function

__metaclass__ = type

from astwro.utils import make_fits_table
from astwro.sampledata import sampledata_dir


def test_fits_list():
    t = make_fits_table(directory=sampledata_dir(), fields=['FILTER', 'EXPTIME', 'OBJECT'], stats=True)
    #t = make_fits_table(directory='/Users/mka/szkola/PhD/iris12bin/input/TZ_For_ND/', fields=['FILTER', 'EXPTIME', 'OBJECT'], stats=True)
    assert (len(t) > 0)



# coding=utf-8
from __future__ import absolute_import, division, print_function
import astropy.io.fits as fits
import astropy.visualization as vis
from astropy.wcs import WCS
import matplotlib.pyplot as plt
from .coord_tools import fix_scamp_wcs
import astwro.sampledata

__metaclass__ = type

def plot_coords(fig=None, ax=None, fits_file=astwro.sampledata.fits_image(), gal=True, img_alpha=1.0, cmap=None,
                grid_c='gray', grid_alpha=0.5, grid_ls='solid', keep_scale=True, img=True,
                ):
    if ax is None or  isinstance(ax, int):
        if fig is None:
            fig = plt.figure(figsize=(6, 6))
            fig.tight_layout()
        if ax is None: ax = 111
        hdu = fits.open(fits_file)[0]
        fix_scamp_wcs(hdu)
        wcs = WCS(hdu.header)
        ax = fig.add_subplot(ax, projection=wcs)
    if img:
        vmin, vmax =  120, 90
        norm = vis.ImageNormalize(vmin=vmin, vmax=vmax, stretch=vis.SqrtStretch())
        ax.imshow(hdu.data, origin='lower', norm=norm, alpha=img_alpha, cmap=cmap)

    ax.coords.grid(True, color=grid_c, ls=grid_ls, alpha=grid_alpha)
    ax.coords[0].set_axislabel('ra')
    ax.coords[1].set_axislabel('dec')
    ax.coords['ra'].set_major_formatter('hh:mm:ss')
    if keep_scale:
        ax.set_autoscale_on(False)
    if gal:
        overlay = ax.get_coords_overlay('galactic')
        overlay.grid(color=grid_c, ls=grid_ls, alpha=grid_alpha)
        overlay[0].set_axislabel('$l$')
        overlay[1].set_axislabel('$b$')
    return ax


# coding=utf-8
from __future__ import absolute_import, division, print_function
#import matplotlib.pyplot as plt



__metaclass__ = type

import numpy as np

from ..fnpeaks import *

class TestFNpeaks:

    def setup_class(self):
        self.N = 10
        self.M = 1000
        self.lc = np.random.normal(10, 0.5, (self.N, self.M))  #STARSxOBSERVATION
        self.lc_e = np.random.normal(0, 0.1, (self.N, self.M))
        self.hjd = 10000.0 + np.sort(np.random.uniform(0, 200, self.M))
        for i in range(self.N):
            self.lc[i] += (1.0 * i / self.N) * np.sin((self.hjd + i)/20.0)

        evens = np.array(np.arange(self.M) & 1, dtype=bool) # divide into two filters ABABAB...
        self.filters = {
            'A': evens,
            'B': ~evens
        }

    def test_simple(self): # find
        runner = FNpeaks(self.hjd, self.lc)
        res = runner(5, periodogram=False) # star == 5
        assert (res is not None)
        assert (len(res.peaks) == 10)
        assert (res.power is None)

    def test_with_periodogram(self):
        runner = FNpeaks(self.hjd, self.lc)
        res = runner(5, periodogram=True)
        assert (res is not None)
        assert (len(res.power) > 0)
        assert (len(res.peaks) == 10)

    def test_filters(self):
        runner = FNpeaks(self.hjd, self.lc, filtermasks=self.filters)
        res = runner(5, periodogram=False, filter='B')
        assert (isinstance(res, dict))
        assert (len(res['B'].peaks) == 10)
        assert (res['B'].power is None)

    def test_filters_errors_periodigrams_ids(self):
        runner = FNpeaks(self.hjd, self.lc, self.lc_e, filtermasks=self.filters, ids=np.arange(self.N)*10)
        res = runner(6, filter='B')
        assert (isinstance(res, dict))
        assert (len(res['B'].peaks) == 10)
        assert (res['B'].power is None)

    def test_fitting(self): # find
        runner = FNpeaks(self.hjd, self.lc)
        res = runner(5, periodogram=False, error_max=1.0, sigmaclip=3) # star == 5
        res.makefit(fitter_class=SineFitter)
        assert (res.fit.amplitude**2 > 0)
        res.makefit(fitter_class=SineFitterC)
        assert (res.fit.amplitude**2 > 0)
        res.makefit(fitter_class=SineHarmonicFitterC)
        assert (res.fit.amplitude**2 > 0)
        assert (res is not None)
        assert (len(res.peaks) == 10)
        assert (res.power is None)



# coding=utf-8
from __future__ import absolute_import, division, print_function
__metaclass__ = type

import os
from .DAORunner import DAORunner
from .OutputProviders import *
from astwro.config import find_opt_file
import astwro.starlist as sl

class Daophot(DAORunner):
    """ **daophot** runner

    Object of this class maintains single process of **daophot** and it's working directory.

    Methods of this class corresponds to **daophot**'s commands, each of those methods
    returns result object providing access to daophot screen output as well as easy access
    to files generated by **daophot** command.
    
    Instance attributes:

    :var str daophotopt:   daophotopt.opt file to be copied
    :var DPOP_OPtion      OPtion_result:     results of command OPtion  or initial options reported by `daophot`
    :var DPOP_ATtach      ATtach_result:     results of command ATtach
    :var DpOp_SKy         SKy_result:        results of command SKy
    :var DpOp_FInd        FInd_result:       results of command FInd
    :var DpOp_PHotometry  PHotometry_result: results of command PHotometry
    :var DpOp_PIck        PIck_result:       results of command PIck
    :var DpOp_PSf         PSf_result:        results of command PSf
    :var DpOp_SOrt        SOrt_result:       results of command SOrt    (not implemented)
    :var DpOp_SUbstar     SUbstar_result:    results of command SUbstar 
    :var str image:  image which will be automatically ATTACHed before every run
    :var options: options which will be automatically added as OPTION command before every run,
                    can be either:
                                dictionary:
                                                        >>> dp = Daophot()
                                                        >>> dp.options = {'GAIN': 9, 'FI': '6.0'}
                                iterable of tuples:
                                                        >>> dp.options = [('GA', 9.0), ('FITTING RADIUS', '6.0')]
                                filename string of daophot.opt-formatted file:
                                                        >>> dp.options = 'config/pydaophot.opt'

    """

    PSF_GAUSSIAN = 1  # type: int
    PSF_MOFFAT15 = 2  # type: int
    PSF_MOFFAT25 = 3  # type: int
    PSF_MOFFAT35 = 4  # type: int
    PSF_LORENTZ = 5   # type: int
    PSF_PENNY1 = 6    # type: int
    PSF_PENNY2 = 7    # type: int

    def __init__(self, dir=None, image=None, daophotopt=None, options=None, batch=False):
        # type: ([str,object], [str], [str], [list,dict], bool) -> Daophot
        """
        :param str dir:          pathname or TmpDir object - working directory for daophot,
                                   if None temp dir will be used and deleted on `Daophot.close()`
        :param str image:        if provided, this file will be automatically attached (AT) as first daophot command
                                   setting image property has same effect
        :param str daophotopt:   daophot.opt file, if None build in default file will be used, can be added later
                                   by `Runner.copy_to_runner_dir(file, 'daophot.opt')`
        :type options: list or str
        :param options: if provided OPTION command will be automatically attached
                                   setting options property has same effect; list of tuples or dict.
                                   Do not set WATCH PROGRESS to sth else than -2
        :param bool batch:         whether Daophot have to work in batch mode.


        .. attribute:: dir

            Runner's directory, object of :class:`astwro.utils.TmpDir`, call ``Daophot.dir.path`` for string path to directory

        """
        if daophotopt:
            self.daophotopt = daophotopt
        else:
            self.daophotopt = find_opt_file('daophot.opt', package='pydaophot')

        self.image = self.expand_path(image)
        self.options = {'WA': -2}
        if options:
            self.options.update(dict(options))

        super(Daophot, self).__init__(dir=dir, batch=batch)
        # base implementation of __init__ calls `_reset` also
        self._update_executable('daophot')

    def _reset(self):
        super(Daophot, self)._reset()

        self.OPtion_result = None  #: Results of command OPtion  or initial options reported by `daophot`
        self.ATtach_result = None
        self.SKy_result = None
        self.FInd_result = None
        self.PHotometry_result = None
        self.PIck_result = None
        self.PSf_result = None
        self.SOrt_result = None
        self.SUbstar_result = None
        self.GRoup_result = None
        self.NEda_result = None


    def __deepcopy__(self, memo):
        from copy import deepcopy
        new = super(Daophot, self).__deepcopy__(memo)

        new.daophotopt = deepcopy(self.daophotopt, memo)

        new.image = deepcopy(self.image, memo)
        new.options = deepcopy(self.options, memo)
        return new

    def _pre_run(self, wait):
        super(Daophot, self)._pre_run(wait)
        if self.options:
            self._enqueueOPtions(self.options, on_beginning=True)
        if self.image:
            self._enqueueATtach(self.expand_path(self.image), on_beginning=True)

        # just for consume options daophot presents on the beginning
        opt_processor = DPOP_OPtion()
        if self.OPtion_result is None:
            self.OPtion_result = opt_processor
        self._insert_processing_step('', output_processor=opt_processor, on_beginning=True)
        # Empty lines to unhang daophot after error (otherwise it waits for corrected input)
        self._insert_processing_step('\n\n\n')

    def _on_exit(self):
        pass


    def _init_workdir_files(self, dir):
        super(Daophot, self)._init_workdir_files(dir)
        self.link_to_runner_dir(self.daophotopt, 'daophot.opt')


    def _enqueueATtach(self, image_file, on_beginning=False):
        # type: (str, bool) -> DPOP_ATtach
        image_file, _ = self._prepare_input_file(image_file)
        processor = DPOP_ATtach()
        self._insert_processing_step('ATTACH {}\n'.format(image_file),
                                     output_processor=processor,
                                     on_beginning = on_beginning)
        if self.ATtach_result is None or not on_beginning:
            self.ATtach_result = processor
        return processor

    def set_options(self, options, value=None):
        # type: ([str,dict,list], [str,float]) -> None
        """
        Set option(s) before run.
        
        Options can be either:

        dictionary:
            ``dp.set_options({'GAIN': 9, 'FI': '6.0'})``

        iterable of tuples:
                                    ``dp.set_options([('GA', 9.0), ('FITTING RADIUS', '6.0')])``

        option key, followed by value in 'value' parameter:
                                    ``dp.set_options('GA', 9.0)``

        filename string of allstar.opt-formatted file (file will be symlinked as `allstar.opt`):
                                    ``dp.set_options('opts/newallstar.opt')``

        .. warning:: Do not set `WATCH PROGRESS` to something else than -2

        :param options: described above
        :param value: value if :attr:`options` is just single key
        :return: results object also accessible as :attr:`Daophot.OPtion_result` property
        :rtype: DPOP_OPtion

        """
        if isinstance(options, str) and value is None:  # filename
            # allstar operates in his tmp dir and has limited buffer for file path
            # so symlink file to its working dir
            self.link_to_runner_dir(options, 'daophot.opt')
        else:
            if self.options is None:
                self.options = {}
            if value is not None:  # single value
                options = {options:value}
            elif isinstance(options, list):
                options = dict(options)
            self.options.update(options)

    # daophot commands
    def ATtach(self, image_file):
        # type: (str) -> DPOP_ATtach
        """
        Add daophot ATTACH command to execution queue. Available only in "batch" mode. 
        
        If image_file parameter is provided in constructor or by set_image method, 
        ATtach is enqueued automatically (preferred method
        until multiple ATTACH commands needed in "batch" mode).
        
        :param str image_file: image to attach file will be symlinked to work dir as ``"i.fits"``,
                   if ``None``, `i.fits` (file or symlink) is expected in working dir
        :return: results object also accessible as :attr:`ATtach_result` property
        :rtype: DPOP_ATtach
        """
        if not self.batch_mode:
            raise Daophot.RunnerException('ATtach is intented for "batch" mode only. Use image property.', self)

        self._get_ready_for_commands()  # wait for completion before changes in working dir

        return self._enqueueATtach(image_file)


    def EXit(self):
        self._insert_processing_step('EXIT\n', output_processor=DaophotCommandOutputProcessor())

    def _enqueueOPtions(self, options, value=None, on_beginning=False):
        # type: ([str, dict, list], [str], bool) -> DPOP_OPtion
        commands = 'OPT\n'
        if isinstance(options, str) and value is None:  # filename
            # daophot operates in his tmp dir and has limited buffer for file path
            # so symlink file to its working dir
            self._get_ready_for_commands()  # wait for completion before changes in working dir
            l_opt, a_opt = self._prepare_input_file(options)
            commands += l_opt+'\n\n'
        else:
            commands += '\n'  # answer for filename
            if value is not None:
                options = [(options,value)]  # options is str with option name and value is it's value
            elif isinstance(options, dict):
                options = options.items()   # options is dict
                                            # else options is list of pairs
            commands += ''.join('%s=%.2f\n' % (k,float(v)) for k,v in options if v is not None)
            commands += '\n'
        processor = DPOP_OPtion()
        self._insert_processing_step(commands, output_processor=processor, on_beginning=on_beginning)
        if self.OPtion_result is None or not on_beginning:
            self.OPtion_result = processor
        return processor


    def OPtions(self, options, value=None):
        # type: ([str, dict, list], [str]) -> DPOP_OPtion
        """
        Adds daophot OPTION command to execution queue. Available only in "batch" mode. 
        
        Use :meth:`set_options()` for options which are set after daophot process start

        Parameter `options` can be either:
                dictionary:
                                        >>> dp = Daophot(mode = "batch")
                                        >>> dp.OPtions({'GAIN': 9, 'FI': '6.0'})
                iterable of tuples:
                                        >>> dp.OPtions([('GA', 9.0), ('FITTING RADIUS', '6.0')])
                option key, followed by value in 'value' parameter:
                                        >>> dp.OPtions('GA', 9.0)
                filename string of daophot.opt-formatted file:
                                        >>> dp.OPtions('config/pydaophot.opt')

        :param options: described above
        :param value: value if `options` is just single key
        :return: results object also accessible as :attr:`Daophot.OPtion_result` property
        :rtype: DPOP_OPtion  
        """
        if not self.batch_mode:
            raise Daophot.RunnerException('OPtions is intented for "batch" mode only. Use set_options().', self)

        self._get_ready_for_commands()  # wait for completion before changes in working dir
        return self._enqueueOPtions(options, value)

    def SKy(self):
        """
        Runs (or adds to execution queue in batch mode) daophot SKY command.

        :return: results object also accessible as :attr:`Daophot.SKy_result` property
        :rtype: DpOp_SKy
        """
        self._get_ready_for_commands()  # wait for completion before changes in working dir
        commands = 'SKY\n'
        processor = DpOp_SKy()
        self._insert_processing_step(commands, output_processor=processor)
        self.SKy_result = processor
        if not self.batch_mode:
            self.run()
        return processor

    def FInd(self, frames_av = 1, frames_sum = 1, starlist_file='i.coo'):
        """
        Runs (or adds to execution queue in batch mode) daophot FIND command.
        
        :param int frames_av: averaged frames in image, default: 1
        :param int frames_sum: summed frames in image, default: 1
        :param str starlist_file: output coo file, default: ``"i.coo"``
        :return: results object also accessible as :attr:`Daophot.FInd_result` property
        :rtype: DpOp_FInd
        """
        self._get_ready_for_commands()  # wait for completion before changes in working dir
        local_starlist_file, abs_starlist_file = self._prepare_output_file(starlist_file)
        commands = 'FIND\n{},{}\n{}\nyes\n'.format(frames_av, frames_sum, local_starlist_file)
        processor = DpOp_FInd(starlist_file=abs_starlist_file)
        self._insert_processing_step(commands, output_processor=processor)
        self.FInd_result = processor
        if not self.batch_mode:
            self.run()
        return processor

    def PHotometry(self, photoopt=None, IS=0, OS=0, apertures=None, stars='i.coo', photometry_file='i.ap'):
        # type: ([str], float, float, [list], [str,sl.StarList], [str]) -> DpOp_PHotometry
        """
        Runs (or adds to execution queue in batch mode) daophot PHOTOMETRY command. 
        
        Either :py:attr:`photoopt` or :attr:`IS`, :attr:`OS` and :attr:`apertures` have to be set.
        eg.: ``IS=35, OS=50, apertures=[8]``

        :param str photoopt: photo.opt file to be used, default: None (provide :attr:`IS`, :attr:`OS` and :attr:`apertures`)
        :param float IS: inner sky radius, overwrites :attr:`photoopt` file value IS
        :param float OS: outer sky radius, overwrites :attr:`photoopt` file value OS
        :param list apertures: apertures radius, up to 12, overwrites :attr:`photoopt` file values A1, A2, ...
        :param  stars: input list of stars, default: ``"i.coo"``
        :type stars: str or StarList
        :param str photometry_file: output magnitudes file
        :return: results object also accessible as `Daophot.PHotometry_result` property
        :rtype: DpOp_PHotometry
        
        .. seealso::  :meth:`NEda`
        """
        # TODO: When PSF file is found, new daophot uses it and behaves differently,
        # TODO: this should avoided, (there is NEda for that),
        # TODO: and not using same name for FITS and PSF by default (avoid finding PSF)
        # TODO: session:
        #           Found PSF file mik.psf
        #       Profile-fitting photometry (default mik.nst): mik.als
        #                     Star ID file (default mik.lst): mik.coo
        #                       Output file (default mik.ap): mik2.ap
        apertures = self._check_apertures(IS, OS, apertures, photoopt)
        self._get_ready_for_commands()  # wait for completion before changes in working dir

        l_popt, a_popt = self._prepare_input_file(photoopt)
        l_star, a_star = self._prepare_input_file(stars, astwro.starlist.DAO.XY_FILE)
        l_phot, a_phot = self._prepare_output_file(photometry_file)

        commands = 'PHOT\n{}\n'.format(l_popt)
        if IS != 0:
            commands += 'IS={}\n'.format(IS)
        if OS != 0:
            commands += 'OS={}\n'.format(OS)
        for i, ap in enumerate(apertures):
            commands += 'A{:X}={}\n'.format(i+1, ap)
        commands += '\n{}\n{}\n'.format(l_star, l_phot)

        processor = DpOp_PHotometry(photometry_file=a_phot)
        self._insert_processing_step(commands, output_processor=processor)
        self.PHotometry_result = processor
        if not self.batch_mode:
            self.run()
        return processor

    def PIck(self, number_of_stars_to_pick=50, faintest_mag=20.0, photometry='i.ap', picked_stars_file='i.lst'):
        # type: (float, float, [str,sl.StarList], str) -> DpOp_PIck
        """
        Runs (or adds to execution queue in batch mode) daophot PICK command. 
        
        :param int: number_of_stars_to_pick
        :param float faintest_mag: instrumental magnitude for the faintest star considered to be picked
        :param photometry:
            input magnitudes file or :class:`~astwro.starlist.StarList`, usually from aperture photometry done by :meth:`PHotometry`.
        :type photometry: str or StarList
        :param str picked_stars_file: output list of picked stars, default: i.lst 
        :return: results object also accessible as :var:`Daophot.PIck_result` property
        :rtype: DpOp_PIck
        """
        self._get_ready_for_commands()  # wait for completion before changes in working dir

        l_phot, a_phot = self._prepare_input_file(photometry)
        l_psfs, a_psfs = self._prepare_output_file(picked_stars_file)

        commands = 'PICK\n{}\n{:d},{:f}\n{}\n'.format(
            l_phot,
            number_of_stars_to_pick,
            faintest_mag,
            l_psfs
        )
        processor = DpOp_PIck(picked_stars_file=a_psfs)
        self._insert_processing_step(commands, output_processor=processor)
        self.PIck_result = processor
        if not self.batch_mode:
            self.run()
        return processor

    def PSf(self, photometry='i.ap', psf_stars='i.lst', psf_file='i.psf'):
        # type: ([str,sl.StarList],  [str,sl.StarList], str) -> DpOp_PSf
        """
        Runs (or adds to execution queue in batch mode) daophot PHOTOMETRY command. 
        
        :param str or sl.StarList photometry: input magnitudes file or Starlist, e.g. from aperture photometry by :func:`PHotometry`.
        :param str or sl.StarList psf_stars: input list of PSF stars, default: i.coo 
        :param str psf_file: output PSF file, default: i.psf
        :return: results object also accessible as `Daophot.PSf_result` property
        :rtype: DpOp_PSf
        """
        self._get_ready_for_commands()  # wait for completion before changes in working dir

        l_phot, a_phot = self._prepare_input_file(photometry)
        l_psfs, a_psfs = self._prepare_input_file(psf_stars)
        l_psf , a_psf  = self._prepare_output_file(psf_file)
        l_nei , a_nei  = self._prepare_output_file(os.path.splitext(l_psf)[0]+'.nei')
        l_err , a_err  = self._prepare_output_file('i.err')

        commands = 'PSF\n{}\n{}\n{}\n'.format(
            l_phot,
            l_psfs,
            l_psf
        )
        processor = DpOp_PSf(psf_file=a_psf, nei_file=a_nei, err_file=a_err)
        self._insert_processing_step(commands, output_processor=processor)
        self.PSf_result = processor
        if not self.batch_mode:
            self.run()
        return processor


    def SOrt(self, file, by, decreasing=None):
        """
        Adds daophot SORT command to execution stack. NOT IMPLEMENTED sorry
        
        Use sorting capabilities of `StarList` and `StarList.renumber()`
        :param str file: fname.COO_FILE etc... any fname.*_FILE to sort
        :param by:  1-based column number, negative for descending order - daophot standard, or
                    one of 'id', 'x', 'y', 'mag'
        :param bool decreasing:  in not None, forces sort order
        :return: results object, also accessible as `Daophot.SOrt_result` property
        :rtype: DpOp_Sort
        """
        self._get_ready_for_commands()  # wait for completion before changes in working dir
        if isinstance(by, str):
            by = by.lower()
            if by == 'id': by = 1
            elif by == 'x': by = 2
            elif by == 'y': by = 3
            elif by == 'mag': by = 4
            else:
                raise Daophot.RunnerValueError('parameter by, if string must be either: "id", "x", "y" or "mag"', self)
        if decreasing is not None:
            by = -abs(by) if decreasing else abs(by)
        raise NotImplementedError("SORT command not implemented")

        if not self.batch_mode:
            self.run()

    def SUbstar(self, subtract, leave_in=None, subtracted_image='is.fits', psf_file='i.psf'):
        # type: (str, str, str, str) -> DpOp_SUbstar
        """
        Adds daophot SUBSTAR command to execution stack.
        
        :param subtract: relative to work dir pathname of stars to subtract file
        :param leave_in: relative to work dir pathname of stars to be kept file (default: None)
        :param psf_file: relative to work dir pathname of file with PSF (default i.psf)
        :param subtracted_image: relative to work dir pathname of output fits file (default is.fits)
        :return: results object, also accessible as `Daophot.SUbstar_result` property
        """
        self._get_ready_for_commands()  # wait for completion before changes in working dir
        l_out, a_out = self._prepare_output_file(subtracted_image)
        l_sub, a_sub = self._prepare_input_file(subtract)
        l_psf, a_psf = self._prepare_input_file(psf_file)
        if leave_in is not None:
            l_lve, a_lve = self._prepare_input_file(leave_in)
            commands = 'SUB\n{}\n{}\ny\n{}\n{}\n'.format(
                l_psf,
                l_sub,
                l_lve,
                l_out
            )
        else:
            commands = 'SUB\n{}\n{}\nn\n{}\n'.format(
                l_psf,
                l_sub,
                l_out
            )
        processor = DpOp_SUbstar(subtracted_image_file=a_out)
        self._insert_processing_step(commands, output_processor=processor)
        self.SUbstar_result = processor
        if not self.batch_mode:
            self.run()
        return processor

    def GRoup(self, photometry='i.ap', psf_file='i.psf', critical_overlap=0.1, groups_file='i.grp'):
        # type: ([str,sl.StarList], str, float, str) -> DpOp_GRoup
        """
        Runs (or adds to execution queue in batch mode) daophot GROUP command to execution stack.
        
        :param str,StarList photometry: stars to be grouped
        :param str psf_file: file with PSF
        :param float critical_overlap: relative to work dir pathname of file with PSF 
        :param str groups_file: output gouped stars file
        :return: results object, also accessible as `GRoup_result` property
        """
        self._get_ready_for_commands()  # wait for completion before changes in working dir
        l_pho, a_pho = self._prepare_input_file(photometry)
        l_psf, a_psf = self._prepare_input_file(psf_file)
        l_grp, a_grp = self._prepare_output_file(groups_file)
        commands = 'GROUP\n{}\n{}\n{}\n{}\n'.format(
            l_pho,
            l_psf,
            critical_overlap,
            l_grp
        )
        processor = DpOp_GRoup(groups_file=a_grp)
        self._insert_processing_step(commands, output_processor=processor)
        self.GRoup_result = processor
        if not self.batch_mode:
            self.run()
        return processor

    def NEda(self, photoopt=None, IS=0, OS=0, apertures=None,
             psf_file='i.psf', psf_photometry='i.als', stars_id='i.als', neda_photometry_file='i.nap'):
        # type: ([str], float, float, [list], [str], [str,sl.StarList], [str,sl.StarList], [str]) -> DpOp_NEda
        """
        Runs (or adds to execution queue in batch mode) daophot NEDA command.

        Performing aperture photometry with neighbours SPF profiles subtraction.
        Either :param photoopt or :param photo_is, :param OS and :param photo_ap have to be set.
        :param [str] photoopt: photo.opt file to be used, default: none 
                    (provide :param photo_is, :param OS and :param photo_ap)
        :param float IS: inner sky radius, overwrites :param photoopt file value IS
        :param float OS: outer sky radius, overwrites :param photoopt file value OS
        :param [list] apertures: apertures radius, up to 12, overwrites `photoopt` file values A1, A2, ...            
        :param str psf_file: file with PSF for profile subtraction
        :param [str, sl.StarList] psf_photometry: stars with PSF photometry for profile subtraction, default: i.als
        :param [str, sl.StarList] stars_id: input list of stars to be measured, default: i.als
        :param str neda_photometry_file: output neda aperture photometry file
        :return: results object, also accessible as `NEda_result` property
        :rtype: DpOp_NEda
        """
        apertures = self._check_apertures(IS, OS, apertures, photoopt)
        self._get_ready_for_commands()  # wait for completion before changes in working dir

        l_popt, a_popt = self._prepare_input_file(photoopt)
        l_psf,  a_psf  = self._prepare_input_file(psf_file)
        l_phot, a_phot = self._prepare_input_file(psf_photometry)
        l_star, a_star = self._prepare_input_file(stars_id, astwro.starlist.DAO.XY_FILE)
        l_neda, a_neda = self._prepare_output_file(neda_photometry_file)

        commands = 'NEDA\n{}\n'.format(l_popt)
        if IS != 0:
            commands += 'IS={}\n'.format(IS)
        if OS != 0:
            commands += 'OS={}\n'.format(OS)
        for i, ap in enumerate(apertures):
            commands += 'A{:X}={}\n'.format(i+1, ap)
        commands += '\n{}\n{}\n{}\n{}\n'.format(
            l_psf,
            l_phot,
            l_star,
            l_neda
        )

        processor = DpOp_NEda(neda_file=a_neda)
        self._insert_processing_step(commands, output_processor=processor)
        self.NEda_result = processor
        if not self.batch_mode:
            self.run()
        return processor

    def _check_apertures(self, IS, OS, apertures, photoopt):
        if photoopt is None and (apertures is None or IS == 0 or OS == 0):
            raise Daophot.RunnerValueError('Apertures and IS and OS must be provided, explicitly or as photoopt file',
                                           self)
        if apertures is None:
            apertures = []
        elif len(apertures) > 12:
            raise Daophot.RunnerValueError('apertures apertures list can contain maximum 12 elements', self)
        return apertures

    def _process_starlist(self, s, **kwargs):
        if kwargs['add_psf_errors']:
            import pandas as pd
            err = self.PSf_result.errors
#            idx = [i for i,_ in err]
#            val = [v for _,v in err]
#            col = pd.Series(val, index=idx)
            new = pd.concat([s, err.loc[:,['psf_err', 'flags']]], axis=1, join_axes=[s.index])
            new.import_metadata(s)
            s = new
        return s




















































# coding=utf-8
from __future__ import absolute_import, division, print_function

__metaclass__ = type

from astwro.phot import stype
import numpy as np

def test_spectral_get_color_values():
    zz = stype.SpectralType({'B-V': 1.5})
    np.testing.assert_allclose(zz.get_color_values('B-V', 'BAFGKM'), [1.2, 1.49, 1.82, 2.1, 2.31, 2.97])

def test_plot_aids():
    zz = stype.SpectralType({'B-V': 1.5})
    aids = zz.generate_plot_aids('B-V')
    print (aids.fill_x)




import glob
from os.path import join, basename
import numpy as np
import astropy.io.fits as fits
from astropy.table import Table
from astropy.stats import SigmaClip


def make_fits_table(directory='.', recursive=False, filename_pattern='*.fits', hud_no=0, fields=None,
                    stats=3.0):
    """
    Returns table with FITS files information

    :param directory: directory to search for FITS files
    :type directory: str, path-like
    :param recursive: **not implemented**
    :type recursive: bool
    :param filename_pattern: glob pattern for fits files
    :type filename_pattern: str
    :param hud_no: Which HUD of FITS file should be scanned for header fields and stats
    :type hud_no: int
    :param fields: list of FITS header fields to be extract to columns
    :type fields: List[str]
    :param stats:   If `not False`, columns with mean, median and std will be added. If numerical, it will be
                    sigma-clipped stats (slower - iterative) with number value indicates value of sigma for clipping.
    :type stats: bool, float
    :rtype Table
    """

    if recursive:
        raise NotImplementedError('recursive search not implemented, ask Mikolaj to implement')
    fits_list = glob.glob(join(directory, filename_pattern))
    if fields is None:
        fields = []
    col_pathn = []
    col_names = []
    cols_fields = [[] for _ in fields]
    cols_stats = [[], [], []] if stats else []
    # cols_stats_names = ['mean', 'median', 'std', 'min', 'max'] if stats else []
    cols_stats_names = ['median', 'min', 'max'] if stats else []
    if isinstance(stats, float):
        clipper = SigmaClip(stats)
    for f in fits_list:
        col_pathn.append(f)
        col_names.append(basename(f))
        if fields or stats:
            with fits.open(f) as hud:
                h = hud[hud_no].header
                for fld, col in zip(fields, cols_fields):
                    try:
                        col.append(h[fld])
                    except LookupError:
                        col.append(None)
                if stats:
                    d = hud[hud_no].data
                    if isinstance(stats, float):
                        clipped, minv, maxv = clipper(d, return_bounds=True)
                        # s = np.nanmean(clipped), np.nanmedian(clipped), np.nanstd(clipped), minv, maxv
                        s = np.nanmedian(clipped), minv, maxv
                    else:
                        # s = np.mean(d), np.median(d), np.std(d), np.min(d), np.max(d)
                        s = np.median(d), np.min(d), np.max(d)
                    for c, v in zip(cols_stats, s):
                        c.append(v)

    return Table([col_pathn, col_names] + cols_stats + cols_fields,
                 names=['pathname', 'filename']+ cols_stats_names + fields)





"""Coordinates calculations.

This package is in early development stage nad can be deeply changed or even excluded

Actually allows image/WCS coordinates calculation using external xt2sky/sky2xy
tools from WCS tools. (which supports sextractor fits header format for distortion)
and matching multiple catalogs at once.
"""

from .coord_tools import *
from .CoordMatch import CoordMatch
from .XY2Sky import XY2Sky
from .Sky2XY import Sky2XY
#from .plots import plot_coords


from astwro.starlist.daofiles import read_dao_file, write_dao_file
from astwro.starlist.ds9 import write_ds9_regions, read_ds9_regions
from astwro.starlist import StarList

COO = 'i.coo'
LST = 'i.lst'
REG = 'i.lst.reg'

# s0 = StarList.new()

coo = read_dao_file(COO)
psf = read_dao_file(LST)
write_ds9_regions(coo, REG, indexes=[psf.index], colors=['red'])
s2 = read_ds9_regions(REG)
print (s2)
print (s2.DAO_hdr)



from .Individual import Individual

class ScoreFunction(object):

    def calc(self, individual):
        """
        Calculates score of individual. To be overridden.
        :param Individual individual:
        :return: float
        """
        return 0


# coding=utf-8
from __future__ import absolute_import, division, print_function
from astropy.coordinates import SkyCoord
import astropy.units as u
import numpy as np


class CoordMatch(object):
    """Two catalogues crossmatch

    Object interface to `astropy.coordinates.match_to_catalog_sky`"""

    def __init__(self, cooA, cooB, radius_match=0.5, radius_separated=None, unit=None):
        super(CoordMatch, self).__init__()
        if radius_separated is None:
            radius_separated = radius_match
        if not isinstance(radius_match, u.Quantity):
            radius_match = radius_match * u.arcsec
        if not isinstance(radius_separated, u.Quantity):
            radius_separated = radius_separated * u.arcsec


        self.r_match = radius_match
        self.r_spe = radius_separated

        kwargs = {} if unit is None else {'unit': unit}
        self.A = SkyCoord(cooA, **kwargs)
        self.B = SkyCoord(cooB, **kwargs)

        self._ABidx = None
        self._ABdist = None

    def _calc_diff(self):
        self._ABidx, self._ABdist, _ = self.A.match_to_catalog_sky(self.B)

    @property
    def sepAB(self):
        if self._ABdist is None:
            self._calc_diff()
        return self._ABdist

    @property
    def mapAB(self):
        if self._ABidx is None:
            self._calc_diff()
        return self._ABidx

    @property
    def lenB(self):
        return len(self.B)

    @property
    def lenA(self):
        return len(self.A)

    @property
    def mAB(self):
        return self.sepAB < self.r_match

    @property
    def mBA(self):
        r = np.zeros_like(self.B, dtype=bool)
        r[self.iBA] = True
        return r

    @property
    def mAonly(self):
        return ~self.mAB

    @property
    def mBonly(self):
        return ~self.mBA


    @property
    def iAonly(self):
        return np.arange(self.lenA)[self.mAonly]

    @property
    def iBonly(self):
        return np.arange(self.lenB)[self.mBonly]

    @property
    def iBA(self):
        return np.unique(self.mappediBA)

    @property
    def iAB(self):
        return np.arange(self.lenA)[self.mAB]

    @property
    def mappediBA(self):
        return self.mapAB[self.mAB]







# coding=utf-8
from __future__ import absolute_import, division, print_function
__metaclass__ = type

import os.path as path
import logging
import logging.handlers
import astropy.io.fits as fits
import astwro.pydaophot
import astwro.starlist

PROGRESS = 25
# Choose number of info dumped to stderr:
# loglevel = logging.DEBUG
# loglevel = logging.INFO
loglevel = PROGRESS
# loglevel = logging.WARNING


# Two loggers: log - stderr and runlog - file logger
logging.addLevelName(PROGRESS, 'PRGRS')
logging.basicConfig(level=loglevel, format='[%(levelname)s] %(message)s')
log = logging.getLogger()
runlog = logging.getLogger('file_runlog')
runlog.propagate = False  # do not pass logrecord to stderr logger ater logging to file


def lstcorr(mult, err, psf, phase):
    averr = err.psf_err.mean()
    err = err[(err.psf_err < mult*averr) & (err.flag == ' ')]  # filter out big errors and * or ? marked stars
    if err.stars_number() != psf.stars_number():
        log.warning('{} PSF stars discarded during {}, error threshold: {}'.format(
            psf.stars_number() - err.stars_number(), phase, mult * averr))
    return psf.loc[err.index]


def daophot_photometry(
        image,     # filename without path
        all_stars, # filename with path
        psf_stars, # filename with path
        IMGPATH = '.', # patch to image
        OUTPATH = '.', # patch for output files
        logfile = 'run.log',
        INPUT = None,
        NPSFV=15,  # minimal nuber of stars for variable PSF
        NPSFC = 5,  # minimal nuber of stars for constant PSF
        ERRPSF1 = 3.0,  # limiting multiplicator for error threshold in calculation of PSF, first iteration
        ERRPSF2 = 2.7,  # limiting multiplicator for error threshold in calculation of PSF, next iterations
        LIMMAGE = 0.25                          	 # limiting magnitude error for output stars [mag]
):
    # setup file logger
    if logfile:
        hdlr = logging.handlers.WatchedFileHandler(logfile)
        frmt = logging.Formatter('%(message)s')
        hdlr.setFormatter(frmt)
        runlog.addHandler(hdlr)
        runlog.setLevel(logging.INFO)
    else:
        runlog.addHandler(logging.NullHandler)

    # check for FITS file existence
    image_filepath = path.join(IMGPATH, image)
    if not path.isfile(image_filepath):
        runlog.error('{}  does not exist', image)
        raise Exception('{} does not exist in directory {}'.format(image, IMGPATH))

    # read headers
    with fits.open(image_filepath) as f:
        f[0].verify('silentfix')
        hdr = f[0].header

    OBSERV = hdr.get('OBSERVAT')
    OBJECT = hdr.get('OBJECT')
    READ_MODE = 2 # hdr.get('GAIN')  # TODO: Where is READ_MODE now??!


    ## set vars here
    if READ_MODE == 2:
        DAOOPT_VAR = 'daophot-Bialkow-2_var.opt'        # name of file to be used as daophot.opt with variable PSF
        DAOOPT_CONST = 'daophot-Bialkow-2_const.opt'    # name of file to be used as daophot.opt with constant PSF
    elif READ_MODE == 16:
        DAOOPT_VAR = 'daophot-Bialkow-16_var.opt'       # name of file to be used as daophot.opt with variable PSF
        DAOOPT_CONST = 'daophot-Bialkow-16_const.opt'   # name of file to be used as daophot.opt with constant PSF
    else:
        raise Exception('Unrecognized read-out mode: {}'.format(READ_MODE))
    ALLOPT = 'allstar-Bialkow.opt'        	  # name of file to be useed as allstar.opt
    PHOOPT = 'photo-Bialkow.opt'                    # name of file to be used as photo.opt


    if not INPUT:
        INPUT = path.join(path.dirname(__file__), 'daophot_bialkow_input')  # for tests only

    log.log(PROGRESS,
        '*************************************\n'
        ' pyDAOPHOT/ALLSTAR workflow for %s\n'
        ' file: %s\n'
        '********************************************\n',
        OBJECT, image
    )

    if not OBSERV:
        OBSERV = 'BIALKOW'     # temporary for testing, OBSERV should be set !

    if OBSERV != 'BIALKOW':
        raise Exception('The header keyword OBSERVAT has value: {}\n'
                        'Wrong observatory name, only BIALKOW permitted\n'.format(image, OBSERV))

    # registering file in run.log

    log.info(image)
    runlog.info(image)
    im_name, _ = path.splitext(image)

    daophot_opt = path.join(INPUT, DAOOPT_VAR)
    allstar_opt = path.join(INPUT, ALLOPT)
    photo_opt   = path.join(INPUT, PHOOPT)

    # Create daophot and allstar wrappers which uses shared temporary directory
    # and running in batch mode - execution of commands on run() method

    sdaophot = astwro.pydaophot.Daophot(image=image_filepath, daophotopt=daophot_opt)
    sallstar = astwro.pydaophot.Allstar(dir=sdaophot.dir, image=image_filepath, allstaropt=allstar_opt)

    # aperture photometry (tr-allps.par)

    sdaophot.PHotometry(photoopt=photo_opt, stars=all_stars)
    # sdaophot.PIck(80,20)
    #idxslst -b=${YSIZE},${XSIZE} -r=${PSFRAD} -e=0.1 i.ap ${im_name}.lst i.lst
    all_ap = sdaophot.PHotometry_result.photometry_starlist
    psf_idx = astwro.starlist.read_dao_file(psf_stars)
    psf_ap = all_ap.loc[psf_idx.index]                         # pandas style select from all_ap which in psf_idx
    psf_ap = psf_ap[(psf_ap.mag > 1.0) & (psf_ap.mag_err < 0.1)] # minmag and maxerr filtering

    if psf_ap.stars_number() < NPSFC:
        raise Exception('There are not enough stars for PSF estimation: {}'.format(psf_ap.stars_number()))

    if psf_ap.stars_number() > NPSFV:
        sdaophot.link_to_runner_dir(path.join(INPUT, DAOOPT_VAR), 'daophot.opt')
    else:
        sdaophot.link_to_runner_dir(path.join(INPUT, DAOOPT_CONST), 'daophot.opt')

    # preliminary estimation of PSF and selection of PSF stars

    sdaophot.PSf(psf_stars=psf_ap)  # psf - i1.par
    # lstcorr -s ${ERRPSF1} i.err i.lst
    psf_ap = lstcorr(ERRPSF1, sdaophot.PSf_result.errors, psf_ap, 'preliminary estimation of PSF')

    if psf_ap.stars_number() < NPSFC:
        raise Exception('There are not enough stars for PSF estimation: {}'.format(psf_ap.stars_number()))

    if psf_ap.stars_number() > NPSFV:
        log.info('Variable PSF model in use: ({})'.format(psf_ap.stars_number()))
    else:
        log.info('Constant PSF model in use: ({})'.format(psf_ap.stars_number()))
        sdaophot.link_to_runner_dir(path.join(INPUT, DAOOPT_CONST), 'daophot.opt')

    # first calculation of PSF

    sdaophot.PSf(psf_stars=psf_ap)  # (psf-i1.par)
    # if (-s i.psf) then
    if not sdaophot.PSf_result.success:
        runlog.error('DAOPHOT cannot create PSF !')
        raise Exception('DAOPHOT cannot create PSF !')
        # lstcorr -s ${ERRPSF1} i.err i.lst
    psf_ap = lstcorr(ERRPSF1, sdaophot.PSf_result.errors, psf_ap, 'first calculation of PSF')
    sallstar.set_options('ma', 100) # psf-subn.par
    sallstar.ALlstar(stars='i.nei')

    # second iteration of PSF

    sdaophot.batch_mode = True  # batch mode allows queuing multiple commands executed together on run() method
    sdaophot.SUbstar(subtract='i.als', leave_in=psf_ap)  # psf-i2.par
    sdaophot.run() # PSf command will delete i.psf do run SUbstar before it happens
    sdaophot.ATtach('is')
    sdaophot.PSf(photometry='i.als', psf_stars=psf_ap)
    sdaophot.run()
    # if (-s i.psf) then
    if not sdaophot.PSf_result.success:
        runlog.error('DAOPHOT cannot create PSF !')
        raise Exception('DAOPHOT cannot create PSF !')
    # lstcorr -s ${ERRPSF2} i.err i.lst
    psf_ap = lstcorr(ERRPSF2, sdaophot.PSf_result.errors, psf_ap, 'second iteration of PSF')
    sallstar.ALlstar(stars='i.nei')  # psf-subn.par

    # third iteration of PSF, grouping

    sdaophot.SUbstar(subtract='i.als', leave_in=psf_ap)  # psf-i2.par
    sdaophot.run() # PSf command will delete i.psf do run SUbstar before it happens
    sdaophot.ATtach('is.fits')
    sdaophot.PSf(photometry='i.als', psf_stars=psf_ap)
    sdaophot.ATtach(image_filepath)
    sdaophot.GRoup(critical_overlap=0.1)
    sdaophot.run()
    # if (-s i.psf) then
    if not sdaophot.PSf_result.success:
        runlog.error('DAOPHOT cannot create PSF !')
        raise Exception('DAOPHOT cannot create PSF !')
    # rmsf 160 i.grp
    # if !(-e i.grp) cp i.ap i.grp
    if sdaophot.GRoup_result.stars < 5:   ## what limit is equiv of rmsf 160 i.grp ?
        sdaophot.copy_to_runner_dir(sdaophot.file_from_runner_dir('i.ap'), 'i.grp')
    hx, hy = sdaophot.PSf_result.hwhm_xy

    # final fit

    sallstar.ALlstar(stars='i.grp')  #phot.par
    # idxecorr ${LIMMAGE} i.als i.tmp
    als = sallstar.ALlstars_result.als_stars
    als = als[als.mag_err < LIMMAGE]
    sdaophot.GRoup(photometry=als, critical_overlap=0.1)  # photgrp.par
    sdaophot.run()
    sallstar.ALlstar(stars='i.grp', subtracted_image_file='is.fits')  #phot.par
    # photsrt.par
    pphot = sallstar.ALlstars_result.als_stars.sort_values('mag')
    #rmsf 200 i.idx
    #if !(-e i.idx) then
    if pphot.stars_number() < 200:  ## what limit is equiv of rmsf 200 i.idx ?
        runlog.error('PSF photometry file not created !')
        raise Exception('PSF photometry file not created!')

    # copying results of PSF photometry

    astwro.starlist.write_dao_file(pphot, path.join(OUTPATH, im_name+'.pphot'))
    sdaophot.copy_from_runner_dir('i.psf', path.join(OUTPATH, im_name+'.psf'))
    sdaophot.copy_from_runner_dir('is.fits', path.join(OUTPATH, im_name+'-sub.fits'))

    #  APERTURE PHOTOMETRY WITH NEDA (2.5 * SEEING)

    A1 = 2.5 * (hx+hy)
    sdaophot.NEda(IS=35, OS=55, apertures=[A1], psf_photometry=pphot, stars_id=pphot)   # neda1.par
    sdaophot.run()
    # if !(-e i.nap) then
    if not sdaophot.NEda_result.success:
        runlog.error('Aperture photometry file not created, NEDA!')
        raise Exception('Aperture photometry file not created, NEDA!')
    aphot = sdaophot.NEda_result.neda_starlist
    # idxecorr ${LIMMAGE} i.ap1 i.tmp
    aphot = aphot[aphot.mag_err < LIMMAGE]
    astwro.starlist.write_dao_file(aphot, path.join(OUTPATH, im_name+'.aphot'), astwro.starlist.DAO.ALS_FILE)

    return sdaophot, sallstar, pphot, aphot, psf_idx, psf_ap



if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description=
                                     " daophot_bialkow for Linux \n"
                                     " by Z. Kolaczkowski, version 2016.10.11 \n"
                                     " rewritten to python/astwro by M. Kaluszynski\n"
                                     " the script makes profile and aperture photometry\n"
                                     " it works properly only for Bialkow/ANDOR CCD frames\n"
                                     " requirements: DAOPHOT/ALLSTAR\n"
                                     " run without arguments works in 'demo' mode on astwro sample data\n",
                                     formatter_class = argparse.RawDescriptionHelpFormatter)
    parser.add_argument('image', default=None, nargs='?', help='filename of FITS file, should be in curr dir')
    parser.add_argument('all_stars', default=None, nargs='?', help='patchname of all stars file')
    parser.add_argument('psf_stars', default=None, nargs='?', help='patchname of PSF stars file')
    arg = parser.parse_args()
    if any([arg.image, arg.all_stars, arg.psf_stars]):
        if not all([arg.image, arg.all_stars, arg.psf_stars]):
            log.error('All: image, all_stars, psf_stars, should be given. Or nothing for "demo" mode.')
            parser.print_usage()
            exit(1)
        impath ='.'
        image = arg.image
        alls = arg.all_stars
        psfs = arg.psf_stars
    else:
        # for testing only
        from astwro.sampledata import fits_image, coo_file, lst_file
        impath, image = path.split(fits_image())
        alls = coo_file()
        psfs = lst_file()

    try:
        r = daophot_photometry(image, all_stars=alls, psf_stars=psfs, IMGPATH=impath)
    except Exception as e:
        log.error('ERROR when processing file {}:\n{} '.format(image, e.message))
        raise





# coding=utf-8
from __future__ import absolute_import, division, print_function

__metaclass__ = type

from astwro.phot import iphas

# def test_iphas():
#     sct = iphas.Simulated_ri_rHa_Colors(reddening=0.44)
#     sct.r_Ha



#! /usr/bin/env python
# coding=utf-8
"""Copy this skeleton to start new tool
"""
from __future__ import print_function, division
# GLOBAL IMPORTS HERE
import __commons as commons
import astwro.starlist as sl
from sys import stdin, stdout, stderr


def __do(arg):
    # TODO: __do arguments - in/out files/streams
    i = stdin
    o = stdout

    # http://stackoverflow.com/questions/14207708/ioerror-errno-32-broken-pipe-python
    from signal import signal, SIGPIPE, SIG_DFL
    signal(SIGPIPE, SIG_DFL)

    if arg.input_format:
        arg.input_format = arg.input_format.upper()
    if arg.output_format:
        arg.output_format = arg.output_format.upper()

    if arg.input_format == 'DS9':
        s = sl.read_ds9_regions(i)
    else:
        daotype = None
        if arg.input_format == 'COO':
            daotype = sl.DAO.COO_FILE
        elif arg.input_format == 'SHORT':
            daotype = sl.DAO.SHORT_FILE
        s = sl.read_dao_file(i, dao_type=daotype)

    if arg.verbose:
        print ('Columns of input file: {}'.format(s.columns), file=stderr)

    if arg.sort is not None:
        s.sort_values(s.columns[arg.sort], ascending=not arg.descending, inplace=True)
        if arg.verbose:
            print ('Sorting by {}'.format(s.columns[arg.sort]), file=stderr)

    if arg.output_format == 'DS9':
        sl.write_ds9_regions(s, o)
    else:
        daotype = sl.DAO.UNKNOWN_FILE
        if arg.output_format == 'COO':
            daotype = sl.DAO.COO_FILE
        elif arg.output_format == 'SHORT':
            daotype = sl.DAO.SHORT_FILE
        sl.write_dao_file(s, o, dao_type=daotype)

    return sl


def __arg_parser():
    import argparse
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=commons.version_string(),
        description='Star List convert\n Converts star list from stdin to desired format on stdout.')
    parser.add_argument('-i', '--input_format',
                        help='input format (default: daophot autodetect), one of:'
                             '\n\tDS9 ds9 region file'
                             '\n\tCOO daophot coo (7 col)'
                             '\n\tSHORT daophot ALS 5 columns')
    parser.add_argument('-o', '--output_format',
                        help='output format (default: same as input), one of:'
                             '\n\tDS9 ds9 region file'
                             '\n\tCOO daophot coo (7 col)'
                             '\n\tSHORT daophot ALS 5 columns')
    parser.add_argument('-s', '--sort', nargs='?', type=int, default=None, const=0, metavar='COL',
                        help='sort output by specified column (default 0)')
    parser.add_argument('-d', '--descending', action='store_true',
                        help='when -s, sort descending (default ascending)')
    parser.add_argument('-V', '--verbose', action='store_true',
                        help='print some info to stderr, eg. detected columns')
    return parser


# Below: standard skeleton for astwro.tools
# CUSTOMIZE:
#   1. customize postional - obligatory arguments
#   2. decide whether to print results (__do return)

def main(**kwargs):
    """Entry point for python script calls. Parameters identical to command line"""
    # Extract default arguments from command line parser and apply kwargs parameters
    args = commons.bunch_kwargs(__arg_parser(), positional=[], **kwargs)
    # call main routine - common form command line and python calls
    return __do(args)


def info():
    commons.info(__arg_parser())


if __name__ == '__main__':
    # Entry point for command line
    __args = __arg_parser().parse_args()  # parse command line arguments
    __do(__args)  # call main routine - common form command line and python calls



# coding=utf-8

import pytest
import glob
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astwro.coord import CoordMatch, central, grouping, match_catalogues, find_duplicates


def teat_test():
    pass



def coo_list(txt):
    return SkyCoord(txt, unit=(u.hourangle, u.deg))

@pytest.fixture
def c1():
    return coo_list( ["1:12:51.68 +1:11:32.3",
                      "1:12:44.0  +1:11:22.6",
                      "1:13:00.1  +1:10:24.5",
                      "1:12:51.7  +1:11:32.0",
                     ])

@pytest.fixture
def c2():
    return coo_list( ["1:14:43.2  +1:12:43",
                      "1:12:51.69 +1:11:32.1",
                      "1:13:00.0  +1:10:24.4",
                      "1:13:00.11 +1:10:24.6",
                      "1:14:44.0  +1:11:22.6",
                     ])


def test_one(c1, c2):
    match = CoordMatch(c1, c2)
    assert match is not None

def test_match(c1, c2):
    match = CoordMatch(c1, c2, 0.5)
    assert (match.iAonly == [1]   ).all()
    assert (match.iAB ==    [0,2,3]   ).all()
    assert (match.iBonly == [0,2,4] ).all()
    assert (match.iBA ==    [1,3]   ).all()


def test_mean(c1):
    c = central(c1)
    assert isinstance(c, SkyCoord)
    assert min(c1.ra) < c.ra < max(c1.ra)
    assert min(c1.dec) < c.dec < max(c1.dec)


@pytest.fixture
def index():    return [3, 2, 3, 2, 3, 1, 7, 6, 9, 8, 9, 10, 9]
@pytest.fixture
def dist():     return [3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  3]
@pytest.fixture
def uniquity(): return [1, 1, 1, 2, 3, 4, 5, 6, 7, 8 ,1 ,1  ,1]

def test_grouping(index, dist):
    grp, lbl, idx = grouping(index, dist, 2.0)
    assert len(grp) == 5
    for n in range(len(index)):
        assert n in grp[idx[n]]

#@pytest.mark.xfail(reason='Not Implemented')
def test_grouping_uniq(index, dist, uniquity):
    grp, lbl, idx = grouping(index, dist, 2.0, labels=uniquity)
    assert len(grp) > 5
    for n in range(len(index)):
        assert n in grp[idx[n]]

def test_matching():
    catalogs = [
        SkyCoord([(10.0, 10.0), (20.0, 20.0), (30.0, 30.0), (30.1, 30.1), (50.0, 50.0), (50.1, 50.1), ], unit=u.deg),
        SkyCoord([(10.1, 10.1), (20.1, 20.1), (30.5, 30.5), (30.6, 30.6), (50.2, 50.2), (50.3, 50.3), ], unit=u.deg),
        SkyCoord([(10.6, 10.6), (89.1, 89.1),  ], unit=u.deg),
        SkyCoord([(10.7, 10.7), (20.3, 20.3),  ], unit=u.deg),
    ]

    c, w, s = match_catalogues(catalogs, 1*u.deg,
                               # ra_col='ra_deg', dec_col='dec_deg', radec_unit=u.deg,
                               # weight_fn=lambda x: x['mag_err']**-2,
                               )
    pass

def test_selfmatching():
    catalog = SkyCoord([(10.0, 10.0), (20.0, 20.0), (30.0, 30.0), (30.1, 30.1), (50.0, 50.0), (50.1, 50.1), ], unit=u.deg)

    mask, coo = find_duplicates(catalog, 1*u.deg)
    newcat = catalog[mask]
    assert len(newcat) == len(coo)


# def test_matching():
#     path = '/Users/michal/projects/TNG/ngc7142/3.1find/'
#     files = glob.glob(f'{path}????0???.vot')
#     catalogs = [Table.read(f) for f in files[:5]]
#
#     c, w, s = match_catalogues(catalogs, 0.5*u.arcsec,
#                                ra_col='ra_deg', dec_col='dec_deg', radec_unit=u.deg,
#                                weight_fn=lambda x: x['mag_err']**-2,
#                                )
#     pass

# def test_grouping_sorting(index, dist):
#     index = [0,1,2,0]
#     dist  = [3,2,1,0]
#
#     grp, idx = grouping(index, dist, 2.0)
#     assert len(grp) == 5
#     for n in range(len(index)):
#         assert n in grp[idx[n]]



import os
import re
import pandas as pd
import astropy.table
from .StarList import StarList

def get_stream(file, mode):
    to_close = []
    if isinstance(file, str):
        file = os.path.expanduser(file)
        f = open(file, mode)
        to_close.append(f)
    else:
        f = file
    return f, to_close


def close_files(to_close):
    for f in to_close:
        f.close()


def extract_SkyCoord(starlist):
    import astropy.units as u
    from astropy.coordinates import SkyCoord

    # type: (StarList) -> SkyCoord

    ra_col = None; ra_unit=None
    dec_col = None; dec_unit = None
    #preferd col names
    for c in ['ra', 'RA', 'RAJ2000']:
        if c in starlist.columns:
            ra_col  = starlist[c]
    for c in ['dec', 'DE', 'de', 'DEC', 'DEJ2000']:
        if c in starlist.columns:
            dec_col  = starlist[c]
    # maybe our deg
    if ra_col is None and dec_col is None and 'ra_deg' in starlist.columns and 'dec_deg' in starlist.columns:
        ra_col  = starlist['ra_deg'];  ra_unit = u.deg
        dec_col = starlist['dec_deg']; dec_unit= u.deg
    else:
        # TODO: regexp
        pass;

    if ra_col is None or dec_col is None:
        return None

    if ra_unit is None:
        try:
            return SkyCoord(ra_col, dec_col)
        except u.UnitsError:
            ra_unit = u.hourangle
            dec_unit = u.deg
    return SkyCoord(ra_col, dec_col, unit=(ra_unit, dec_unit))


def as_starlist(input, read_files=True, raise_exceptions=True, updateskycoord=True):
    if isinstance(input, StarList):
        return input

    # TODO: input:Path -> string
    # TODO: input:string : read file: DAO, astropy, pandas
    if isinstance(input, astropy.table.Table):
        meta=input.meta
        input = StarList(input.to_pandas())
        input.DAO_type = meta.get('DAO_type')
        dao_hdr = {}
        for key in ['NL', 'NX', 'NY', 'LOWBAD', 'HIGHBAD', 'THRESH', 'AP1', 'PH/ADU', 'RNOISE', 'FRAD']:
            v = meta.get(key)
            if v is not None:
                dao_hdr[key] = v
        if len(dao_hdr) > 0:
            input.DAO_hdr = dao_hdr

    elif isinstance(input, pd.DataFrame):
        input = StarList(input)
    # TODO: structured numpy arrays
    elif raise_exceptions:
        raise TypeError('Cannot interpret type {} as astropy.starlist.StarList'.format(type(input)))
    else:
        return None

    # bytes->string
    for col in input:
        if isinstance(input[col][0], bytes):
            input[col] = input[col].str.decode('utf-8')

    # check id obligatory field/idx
    input.refresh_id()
    # check, update sky ra/dec coords
    if updateskycoord:
        coo = extract_SkyCoord(input)
        if coo:
            input.radec_hmsdms_from_skycoord(coo)
            input.radec_deg_from_hmsdms()
    return input




# coding=utf-8
from __future__ import absolute_import, division, print_function

from astropy.coordinates import Angle, SkyCoord
import astropy.units as u
from astropy.table import Table
import numpy as np
from os import path

import logging
logger = logging.getLogger(__name__)
#logger.addHandler(logging.NullHandler())


def central(coords, weights=None, unit=None, axis=None):
    """ Calculates coordinates of the center of provided set of coordinates
    """
    try:
        if coords.ndim > 1:
            return _ndimcentral(coords, weights=weights, unit=unit, axies=axis)
    except AttributeError:
        pass
    return  _1dimcentral(coords, weights=weights, unit=unit)

def box(coords, unit=None, expand=True):
    """ Box (rectangle) containing all the `coords`

    Returns
    -------
    (center:SkyCoord, ra_size:Angle, dec_size:Angle)
    """
    unit_kwargs = {}
    if unit is not None:
        unit_kwargs['unit'] = unit
    if isinstance(coords, Table):
        try:
            try:
                coords = SkyCoord.guess_from_table(coords[['ra', 'dec']], **unit_kwargs)
            except u.UnitsError:
                coords = SkyCoord.guess_from_table(coords[['ra', 'dec']], unit=(u.hourangle, u.deg))
        except (KeyError, AttributeError):
            coords = SkyCoord.guess_from_table(coords, **unit_kwargs)
    else:
        coords = SkyCoord(coords, **unit_kwargs)

    dra = coords.ra.max() - coords.ra.min()
    ddec = coords.dec.max() - coords.dec.min()
    cra = coords.ra.min() + dra / 2.0
    cdec = coords.dec.min() + ddec / 2.0
    if expand:
        if isinstance(expand, bool):
            expand = 1.1
        dra  *= expand
        ddec *= expand

    return SkyCoord(cra, cdec), dra, ddec


def _1dimcentral(coords, weights, unit):

    if unit is not None:
        coords = SkyCoord(coords, unit=unit)

    ref_coords = coords[0]

    ra_offsets, dec_offsets = ref_coords.spherical_offsets_to(coords)
    if weights is not None:
        ra_offset  =  sum(ra_offsets  * weights) / sum(weights)
        dec_offset =  sum(dec_offsets * weights) / sum(weights)
    else:
        ra_offset  =  ra_offsets.mean()
        dec_offset =  dec_offsets.mean()

    return SkyCoord(ref_coords.ra + ra_offset, ref_coords.dec + dec_offset)



def _ndimcentral(coords, weights, unit, axies):
    raise NotImplementedError('vectorized central not implemented yet')
    ##_ndimcentral = np.vectorize(_1dimcentral, otypes=[SkyCoord], excluded=['unit'])





def _prepare_labels(labels_iter):
    if not labels_iter: # none or empty
        return None
    if isinstance(labels_iter[0], set): # sets are ok
        return labels_iter
    return [{e} for e in labels_iter] # convert fot list os one-element sets


def grouping(index, dist, radius, labels=None, unique_labels=True):
    """
    Finds groups of neighbour objects.

    `index` and `dist` like returned by selfmatchig
    by `astropy.coordinates.SkyCoord.match_to_catalog_sky` ::

        index, dist = coo.match_to_catalog_sky(coo, 2)

    Each element will belong to group with closest neighbor if distance to it < `radius`.
    (This does no guaranties that members of different groups are distant by more than `radius`
    in the general case.)

    Parameters
    ----------
    index : array_like
        Index of closest neighbours
    dist : array_like
        Distances to neighbours from index
    radius : float or Angle
        Radius of grouping
    labels : array_like, optional
        Array of sets. Optional labels for elements. Sets of labels will be returned.
    unique_labels : bool, optional
        If `true`, two elements can be grouped only when they labels sets are disjoint. E.g. ::

            uniquity = [{1,2}, {3,4,5}, {1,6}]

        element 0 cannot be grouped with element 2 because of label `1`.
    Returns
    -------
    groups : list of sets
        List of groups - sets of group members
    grlabels: list of sets
        List of group labels - sets of group labels
    assignation : list
        Index of group assignation `assignation[4]` is a name of group containing 4th element
        `n in groups[assignation[n]] == True`

    """
    labels = _prepare_labels(labels)
    if labels is None:
        unique_labels = False
    groups = {}  # indexed by group names (id of leaders), values - ids of members
    grlabels = {} # indexed by group names (id of leaders), values - labels of members
    assigned = {}  # indexed by ids, values of group names

    it = np.argsort(dist) if labels else range(len(index))  # w/o labels check sorting has no effect on result
    for n in it:
        i = n
        grp = set(); lbl = set()
        while i not in assigned and i not in grp:
            grp.add(i)
            try:
                lbl.update(labels[i])
            except: pass
            if dist[i] < radius: # progress
                if not unique_labels:
                    i = index[i]
                else:  # check labels before progressing
                    nxt = index[i]
                    try:
                        ilabels = grlabels[assigned[nxt]] # colect group labels
                    except KeyError:
                        ilabels = labels[nxt]   # or personal labels
                    if lbl.isdisjoint(ilabels):
                        i = nxt
        if grp:
            try:
                grp_name = assigned[i]
                groups[grp_name].update(grp)
                grlabels[grp_name].update(lbl)
            except KeyError:
                grp_name = n  # new group
                groups[grp_name] = grp
                grlabels[grp_name] = lbl
            for i in grp:
                assigned[i] = grp_name
    assert (len(assigned) == len(index))
    # dict to list, three dictionaries groups, grlabels and assigned are converted to lists
    grp_names = groups.keys()
    translate = {grname: pos for pos, grname in enumerate(grp_names)}
    return [groups[k] for k in grp_names], [grlabels[k] for k in grp_names],[translate[assigned[i]] for i in range(len(assigned))]


def find_duplicates(catalogue, radius, coord_col=None, ra_col=None, dec_col=None, radec_unit=None, weight_fn=None):
    """
    Finds close neighbours

   Usage ::

        mask, new_coo = find_duplicates(star_list, 0.5*u.arcsec)
        new_starlist = star_list[mask]
        new_starlist['ra_deg'] = new_coo.ra
        new_starlist['dec_deg'] = new_coo.dec

    Parameters
    ----------
    catalogue : SkyCoord  or array_like or StarList
        Catalogue
    radius : float or Quantity
        Matching radius
    ra_col : int or str, optional
        Index of catalog column for RA (only if catalogs are not SkyCoord).
    dec_col : int or str, optional
        Index of catalog column for DEC (only if catalogs are not SkyCoord).
    coord_col : int or str, optional
        Index of catalog column of type SkyCoord (only if catalogs are `astropy.table.Table` with SkyCoord column).
    radec_unit : astropy.unit, optional
        Unit for ra,dec cols conversion: `astropy.coordinates.SkyCoord(cat[ra_col], cat[dec_col], unit=radec_unit)`
    weight_fn : callable, optional
        Function returning weight for given catalogue element. Default: `lambda x: 1.0`

    Returns
    -------
    coords : SkyCoord
        Coordinates of matched objects
    weights: list
        Summed weights for groups
    src : list of sets
        Source catalogs for matched objects - sets of catalog indexes

    """

    logger.info('Starting self-matching catalogue'.format(len(catalogue)))
    # weighting  (calc weight for element of joined catalog)
    if weight_fn is None:
        weights = [1.0] * len(catalogue)
    else:
        weights = [weight_fn(r) for r in catalogue]
        logger.info('Weigths applied')

    # prepare coords
    if coord_col:
        assert ra_col is None and dec_col is None, 'Either coord_col or (ra_col,dec_col) can be specified but not both'
        coords = catalogue[coord_col]
    elif ra_col or dec_col:
        assert ra_col and dec_col, 'Both or none of ra_col and dec_col can be specified'
        skykwargs = {'unit': radec_unit} if radec_unit is not None else {}
        coords = SkyCoord(catalogue[ra_col], catalogue[dec_col], **skykwargs)
    else:
        try:
            coords = catalogue.to_skycoords()  # StarList?
        except AttributeError:
            coords = catalogue

    c = coords
    w = np.array(weights)

    # calc distances
    idx, dis, _ = coords.match_to_catalog_sky(coords, nthneighbor=2)
    # group
    grp, lbl, assgn = grouping(idx, dis, radius=radius, unique_labels=False)
    mask = [i == min(grp[ass]) for i, ass in enumerate(assgn)]
    logger.info('{:6d} objects left'.format(len(grp)))

    # calc coord and weights for groups
    c = SkyCoord([mean_coord(c[list(g)], w[list(g)]) for g in grp], unit=u.deg)

    return mask, c



def match_catalogues(catalogues, radius, selfmatch_radius=0.0, max_iter=200,
                     coord_col=None, ra_col=None, dec_col=None, radec_unit=None, weight_fn=None):
    """
    Matches the objects catalogues.


    Parameters
    ----------
    catalogues : list
        List of catalogues. Catalogue can be SkyCoord object, or array_like with coord columns.
    radius : float or Quantity
        Matching radius
    max_iter : int, optional
        Maximum number of iterations
    selfmatch_radius : float or array_like, optional
        If given, self-matching with that radius performed without preserving distinct detection from one catalogue.
        Use in the case of double detections of deformed objects can occur in any of the catalogues.
    ra_col : int or str, optional
        Index of catalog column for RA (only if catalogs are not SkyCoord).
    dec_col : int or str, optional
        Index of catalog column for DEC (only if catalogs are not SkyCoord).
    coord_col : int or str, optional
        Index of catalog column of type SkyCoord (only if catalogs are `astropy.table.Table` with SkyCoord column).
    radec_unit : astropy.unit, optional
        Unit for ra,dec cols conversion: `astropy.coordinates.SkyCoord(cat[ra_col], cat[dec_col], unit=radec_unit)`
    weight_fn : callable, optional
        Function returning weight for given catalogue element. Default: `lambda x: 1.0`

    Returns
    -------
    coords : SkyCoord
        Coordinates of matched objects
    weights: list
        Summed weights for groups
    src : list of sets
        Source catalogs for matched objects - sets of catalog indexes

    """


    # labeling (assigning each catalog a number-label, storing that label for each element of joined catalog)
    labels = []
    for i,c in enumerate(catalogues):
        labels += [i]*len(c)

    logger.info('Starting matching {} catalogues'.format(len(catalogues)))
    # weighting  (calc weight for element of joined catalog)
    if weight_fn is None:
        weights = np.ones_like(labels)
    else:
        weights = []
        for i, c in enumerate(catalogues):
            weights += [weight_fn(r) for r in c]
        logger.info('Weigths applied')

    # prepare coords
    if coord_col:
        assert ra_col is None and dec_col is None, 'Either coord_col or (ra_col,dec_col) can be specified but not both'
        coord_list = [cat[coord_col] for cat in catalogues]
    elif ra_col or dec_col:
        assert ra_col and dec_col, 'Both or none of ra_col and dec_col can be specified'
        skykwargs = {'unit': radec_unit} if radec_unit is not None else {}
        coord_list = [SkyCoord(cat[ra_col], cat[dec_col], **skykwargs) for cat in catalogues]
    else:
        coord_list = catalogues
    # stack up
    coords = SkyCoord(coord_list)
    logger.info('Joined catalog of {} objects created'.format(len(coords)))
    logger.info('Matching catalogs with radius {}'.format(radius))

    c = coords
    l = labels
    w = np.array(weights)
    neighbour_order = 2
    neighbour_order_increased = False
    selfmatch_performed = False
    for iteration in range(max_iter):
        # calc distances
        idx, dis, _ = c.match_to_catalog_sky(c, neighbour_order)
        # group
        grp, lbl, assgn = grouping(idx, dis, radius=radius, labels=l, unique_labels=not selfmatch_performed)
        reduced = len(c) - len(grp)
        assert reduced >= 0
        logger.info('{:4d} iteration, order {}: {:6d} objects left'.format(iteration, neighbour_order, len(grp)))
        if reduced == 0:
            if not neighbour_order_increased:
                neighbour_order += 1
                neighbour_order_increased = True
            elif not selfmatch_performed and selfmatch_radius:
                neighbour_order = 2
                radius = selfmatch_radius
                selfmatch_performed = True
                logger.info('Self-match with radius {}'.format(selfmatch_radius))
            else:
                break
        else:
            neighbour_order_increased = False
        # calc coord and weights for groups
        c = SkyCoord([mean_coord(c[list(g)], w[list(g)]) for g in grp], unit=u.deg)
        w = np.array([sum(w[list(g)]) for g in grp])
        l = lbl

    logger.info('Matchng finished {:6d} of {:6d} objects left'.format(len(c), len(coords)))
    return c,w,l




def mean_coord(sky_coords, weigths=None):
    if isinstance(mean_coord, SkyCoord):  ## TODO: something is wrong here!
        coo = np.array([sky_coords.ra.ra, sky_coords.ra.deg])
    else:
        coo = np.array([[s.ra.deg, s.dec.deg] for s in sky_coords]).T
    if weigths is not None:
        ws = sum(weigths)
        coo = coo * weigths
    else:
        ws = len(sky_coords)

    ra = coo[0].sum() / ws
    dec = coo[1].sum() / ws
    return ra, dec


# def self_match(coords: SkyCoord, radius: Angle, weights=None):
#     # type: (SkyCoord, Angle, array_like) -> object
#     if weights is None:
#         weights = np.ones_like()
#     c = coords
#     n = 2
#     while n <= 2 or   min(dist) < radius:
#         idx, dist = c.match_to_catalog_sky()


def xy2sky(x, y, transformer, method='try'):
    """Uses external `xy2sky` tool """
    from astwro.coord import XY2Sky
    return XY2Sky(transformer)(x,y)

def skyradec2xy(ra, dec, unit='deg', transformer=None, method='try'):
    """Uses external `xy2sky` tool """
    from astwro.coord import Sky2XY
    return Sky2XY(transformer)(ra=ra, dec=dec, unit=unit)

def skycoo2xy(coo, transformer, method='try'):
    """Uses external `xy2sky` tool """
    from astwro.coord import Sky2XY
    return Sky2XY(transformer)(coo=coo)


# def xy2sky(x, y, transformer, method='try'):
#     from subprocess import Popen, PIPE
#     x = np.atleast_1d(x)
#     y = np.atleast_1d(y)
#     from astwro.utils import TmpDir
#     with TmpDir() as d:
#         xyfile = path.join(d.path, 'xy')
#         np.savetxt(xyfile, (x,y))
#         # xy.tofile(xyfile, sep=' ')
#         # with open(xyfile, 'a') as f:  # new line ath the end
#         #     f.write('\n')
#         try:
#             p = Popen(['xy2sky', '-dn', '6', transformer, '@'+xyfile], stdout=PIPE, stderr=PIPE)
#             output, err = p.communicate()
#         except FileNotFoundError:
#             raise FileNotFoundError('xy2sky from WCSTools must be installed')
#         pass
#     if p.returncode != 0:
#         raise Exception('xy2sky failed:\n'+err)
#     a = np.loadtxt(output.raw)
#     pass
#
#


def fix_scamp_wcs(hdu):
    """Make FITS image scamp WCS headers WCSLib compatible

    Warning
    -------
    Current implementation just removes distortion parameters,
    astrometry is not accurate anymore!
    """
    for i in range(11):
        for pv in ['PV1_{:d}', 'PV2_{:d}']:
            hdu.header.remove(pv.format(i), ignore_missing=True, remove_all=True)







# coding=utf-8
from __future__ import absolute_import, division, print_function
from astropy.coordinates import SkyCoord
import astropy.units as u
import numpy as np
from os import path
from io import StringIO
from .WCSUtilsRunner import WCSUtilsRunner



class Sky2XY(WCSUtilsRunner):
    def __init__(self, translator=None):
        # base implementation of __init__ calls `_reset` also
        super(Sky2XY, self).__init__(translator=translator)
        self._update_executable('sky2xy')

    def __call__(self, coo=None, ra=None, dec=None, unit=None, translator=None):
        """
        Parameters
        ----------
        coo : SkyCoord
        ra,dec : float or array_like or Angle or Quantity
        unit : str or Unit
            default deg


        Returns
        -------
        pixel : numpy.array
        """
        if coo is not None:
            assert ra is None and dec is None and unit is None
        else:
            assert ra is not None and dec is not None
            if unit is not None and unit != 'deg' and unit != u.deg:
                coo = SkyCoord(ra, dec, unit=unit)

        if coo is not None:
            ra = coo.ra.deg
            dec = coo.dec.deg

        if translator is None:
            translator = self.translator

        assert translator is not None, 'Set fits file with WCS as "translator"'

        translator = path.abspath(translator)
        self._prepare_input_file(translator, preservefilename = True)
        ra = np.atleast_1d(ra)
        dec = np.atleast_1d(dec)
        xyfile = path.join(self.dir.path, 'radec')
        np.savetxt(xyfile, np.column_stack([ra,dec]), fmt='%12.6f')


        self.arguments = ['-jn', '3', translator, '@radec']
        self.run()
        return  np.loadtxt(StringIO(self.output), usecols=[4,5], unpack=True)



# coding=utf-8
from __future__ import absolute_import, division, print_function
from logging import *

__metaclass__ = type




class AbstractOutputProvider(object):
    # Abstract class (interface) for chained stream processing
    def get_output_stream(self):
        """To be overridden"""
        pass


class StreamKeeper(AbstractOutputProvider):
    """First in the chain, just keeps stream handler"""
    stream = None
    runner = None

    def __init__(self, runner=None, stream=None):
        super(StreamKeeper, self).__init__()
        self.stream = stream
        self.runner = runner

    def get_output_stream(self):
        if self.stream is None:
            raise Exception('No output stream available, call run() before collecting results.')
        return self.stream


class OutputProvider(AbstractOutputProvider):
    # Base class for elements of stream processors chain
    #    also can be used as dummy processor in chain

    def __init__(self, prev_in_chain=None):
        self.__stream = None
        self._prev_in_chain = prev_in_chain # previous output provider
        self.logger = getLogger('ResultParser')

    def _consume(self, stream):
        """ to be overridden
        :param stream:
        """
        pass

    def get_output_stream(self):
        if self.__stream is None:
            self.__stream = self._prev_in_chain.get_output_stream()
            self._consume(self.__stream)
        return self.__stream

    @property
    def success(self):
        """True if command succeed and output is ready """
        try:
            self.raise_if_error()
        except:
            return False
        else:
            return True

    def raise_if_error(self):
        """ Should raise exception if not properly processed:
            - command did not run
            - buffer analysis indicates error
            - no output value found...
            User can call it to check if command was successful
            To be overridden """
        pass




class OutputLinesProcessor(OutputProvider):
    # Base for line-by-line processing

    def _process_line(self, line, counter):
        """ processes line by line output
            return True if it's last line.
            To be overridden """
        return True

    def _consume(self, stream):
        counter = 0
        for line in stream:
            counter += 1
            self.logger.debug("Output line %3d: %s", counter, line)
            last_one = self._process_line(line, counter)
            if last_one:
                self.logger.debug("Was last line")
                return


class OutputBufferedProcessor(OutputLinesProcessor):
    __buffer = ''

    def _process_line(self, line, counter):
        """ processes line-by-line output
            return True if it's last line.
            If overridden, this base impl should be called """
        self.__buffer += line
        return self._is_last_one(line, counter)

    def _is_last_one(self, line, counter):
        """ return True if it's last line.
            To be overridden """

    def get_buffer(self):
        self.get_output_stream()  # tigers processing
        self.logger.debug("Buffer of output obtained: " + self.__buffer)
        return self.__buffer




# coding=utf-8
from astwro.sampledata import  fits_image
from astwro.coord import xy2sky, skycoo2xy, skyradec2xy
import astropy.units as u


def test_xy2sky():
    coo1 = xy2sky(100,121,fits_image())
    coo2 = xy2sky([100,100], [120,121], fits_image())

    assert coo1.ra == coo2[1].ra
    assert coo1.ra > 301*u.deg


def test_sky2xy():
    coo1 = xy2sky(100,100, transformer=fits_image())
    coo2 = xy2sky([100,100, 99], [120,121,12.8], transformer=fits_image())
    xy1 = skycoo2xy(coo1, transformer=fits_image())
    xy2 = skycoo2xy(coo2, transformer=fits_image())

    assert xy2[1,2] < 13.0 and xy2[1,2] > 12.0


def test_long_tmp():
    ra = [10, 10, 9]
    dec = [12, 21, 12.8]
    packed = skyradec2xy(ra, dec, unit=(u.deg, u.deg), transformer=fits_image())
    s = packed.shape
    x, y = packed



import os


class CycleFile(object):
    """
    Creates series of files with names containing counter, with symlink to newest one
    """

    def __init__(self, path, basename, extension='', create_symlinks=True, symlink_suffix='_last', auto_close=True):
        """
        :param auto_close:
        :param str path: patch to the file, absolute or relative, can be empty string '' for current directory
        :param str basename: first part of file name before counter
        :param str extension:  part of filename after counter
        :param str create_symlinks: whether to create symlink to newest file
        :param str symlink_suffix:  part of suffix filename in place of counter value
        :param bool auto_close: close prev files on next_file nad destruction
        """
        self.path = path
        self.basename = basename
        self.extension = extension
        self.create_symlinks = create_symlinks
        self.symlink_suffix = symlink_suffix
        self.auto_close = auto_close
        self.counter = 0
        self.file = None
        self.filename_formatter = '{}{:03d}{}'
        self.symlink_formatter = '{}{}{}'
        self.file_mode = 'w'
        self.open_kwargs = None
        self.filepath = None
        pass

    def __enter__(self):
        return self
    def __exit__(self, type, value, traceback):
        self.close()
        return True

    def __del__(self):
        if self.auto_close and self.file is not None:
            self.close()

    def next_file(self, counter=None):
        if counter is None:
            counter = self.counter + 1
        self.counter = counter
        if self.auto_close and self.file is not None:
            self.file.close()
            self.file = None
        self._link()
        filename = self.filename_formatter.format(self.basename, counter, self.extension)
        filepath = os.path.join(self.path, filename)
        if self.open_kwargs is not None:
            self.file = open(filepath, self.file_mode, **self.open_kwargs)
        else:
            self.file = open(filepath, self.file_mode)
        self.filepath = filepath
        return self.file


    def close(self):
        if self.file is not None:
            self.file.close()
        self._link()


    def _link(self):
        if self.create_symlinks and self.filepath:
            symlname = self.symlink_formatter.format(self.basename, self.symlink_suffix, self.extension)
            symlpath = os.path.join(self.path, symlname)
            if os.path.exists(symlpath):
                os.remove(symlpath)
            os.symlink(self.filepath, symlpath)



from __future__ import absolute_import, division, print_function
__metaclass__ = type
# see https://wiki.python.org/moin/PortingToPy3k/BilingualQuickRef
from .DAORunner import DAORunner
from .OutputProviders import *
from astwro.config import find_opt_file
import astwro.starlist


class Allstar(DAORunner):
    """ `daophot` runner

    Object of this class maintains single process of `allstar` and it's working directory.

    Instance attributes:
    :var str allstaropt:   allstar.opt file to be copied into runner dir
    :var DPOP_OPtion  OPtion_result:     initial options reported by `allstar`
    :var DPOP_ATtach  ATtach_result:     results of command ATtach
    :var str image:  image which will be automatically used if not provided in `ALlstars` command
    :var dict options: options which will be automatically set as OPTION command before every run,
                    can be either:
                                dictionary:
                                                        >>> dp = Allstar()
                                                        >>> dp.options = {'PROFILE ERROR': 5, 'FI': '6.0'}
                                iterable of tuples:
                                                        >>> dp.options = [('PR', 5.0), ('FITTING RADIUS', 6.0)]
    """

    def __init__(self, dir=None, image=None, allstaropt=None, options=None, batch=False):
        # type: ([str,object], [str], [str], [list,dict], bool) -> Allstar
        """
        :param [str] dir:          pathname or TmpDir object - working directory for daophot,
                                   if None temp dir will be used and deleted on `Allstar.close()`
        :param [str] image:        if provided this file will be used if not provided in `ALlstars` command
                                   setting image property has same effect
        :param [str] allstaropt:   allstar.opt file, if None build in default file will be used, can be added later
                                   by `Runner.copy_to_runner_dir(file, 'allstar.opt')`
        :param [list,dict] options: if provided options will be set on beginning of each process
                                   list of tuples or dict
        :param bool batch:         whether Allstar have to work in batch mode. 
        """
        if allstaropt is not None:
            self.allstaropt = allstaropt
        else:
            self.allstaropt = find_opt_file('allstar.opt', package='pydaophot')

        self.image = self.expand_path(image)

        self.options = {'WA': 0}  # suppress animations
        if options:
            self.options.update(dict(options))

        super(Allstar, self).__init__(dir=dir, batch=batch)
        # base implementation of __init__ calls `_reset` also
        self._update_executable('allstar')

    def _reset(self):
        super(Allstar, self)._reset()
        self.OPtion_result = None
        self.ALlstars_result = None


    def __deepcopy__(self, memo):
        from copy import deepcopy
        new = super(Allstar, self).__deepcopy__(memo)

        new.allstaropt = deepcopy(self.allstaropt, memo)

        # new.OPtion_result = deepcopy(self.OPtion_result, memo)
        # new.ALlstars_result = deepcopy(self.ALlstars_result, memo)

        new.image = deepcopy(self.image, memo)
        new.options = deepcopy(self.options, memo)
        return new

    def _pre_run(self, wait):
        if not self.ALlstars_result:
            raise Allstar.RunnerException('Add ALlstar command before run.', self)
        super(Allstar, self)._pre_run(wait)
        # set options, and prepare options parser
        commands = ''
        if self.options:  # set options before
            options = self.options
            if isinstance(options, dict):
                options = options.items()  # options is dict
            # else options is list of pairs
            commands += ''.join('%s=%.2f\n' % (k, float(v)) for k, v in options if v is not None)
        commands += '\n'
        processor = AsOp_opt()
        self.OPtion_result = processor
        self._insert_processing_step(commands, output_processor=processor, on_beginning=True)

    def _init_workdir_files(self, dir):
        super(Allstar, self)._init_workdir_files(dir)
        self.link_to_runner_dir(self.allstaropt, 'allstar.opt')

    def set_options(self, options, value=None):
        # type: ([str,dict,list], [str,float]) -> None
        """set option(s) before run. 
        
        Options can be either:
                dictionary:             `dp.OPtion({'GAIN': 9, 'FI': '6.0'})`
                iterable of tuples:     `dp.OPtion([('GA', 9.0), ('FITTING RADIUS', '6.0')])`
                option key, followed by value in 'value' parameter:
                                        `dp.OPtion('GA', 9.0)`
                filename string of allstar.opt-formatted file (file will be symlinked as `allstar.opt`):
                                        `dp.OPtion('opts/newallstar.opt')`
        Once set, options will stay set in next runs, set option to `None` to unset
                """
        if isinstance(options, str) and value is None:  # filename
            # allstar operates in his tmp dir and has limited buffer for file path
            # so symlink file to its working dir
            self.link_to_runner_dir(options, 'allstar.opt')
        else:
            if self.options is None:
                self.options = {}
            if value is not None:  # single value
                options = {options:value}
            elif isinstance(options, list):
                options = dict(options)
            self.options.update(options)

    def ALlstar(self, image_file=None, psf_file='i.psf', stars='i.ap', profile_photometry_file='i.als', subtracted_image_file=None):
        # type: ([str], str, [str,object], str, str) -> AsOp_result
        """
        Runs (or adds to execution queue in batch mode) daophot PICK command. 
        :param [str] image_file: input image filepath, if None, one set in constructor or 'i.fits' will be used
        :param str psf_file: input file with psf from daophot PSF command
        :param str stars: input magnitudes file, e.g. from aperture photometry done by :func:`Daophot.PHotometry`.
        :param str profile_photometry_file: output file with aperture photometry results, default: i.als 
        :param str subtracted_image_file: output file with subtracted FITS image, default: do not generate image
        :return: results object also accessible as :var:`Allstar.ALlstars_result` property
        :rtype: AsOp_result
        """
        self._get_ready_for_commands()  # wait for completion before changes in working dir

        if not image_file:
            image_file = self.image
        if not image_file:
            image_file = 'i.fits'

        l_img, a_img = self._prepare_input_file(image_file)
        l_psf, a_psf = self._prepare_input_file(psf_file)
        l_pht, a_pht = self._prepare_input_file(stars, astwro.starlist.DAO.XY_FILE)
        l_als, a_als = self._prepare_output_file(profile_photometry_file)
        l_sub, a_sub = self._prepare_output_file(subtracted_image_file)

        commands = '{}\n{}\n{}\n{}\n{}'.format(l_img, l_psf, l_pht, l_als, l_sub)
        if l_sub:
            commands += '\n'  # if subtracted image not needed, EOF (without new line) should be answer for it (allstar)

        processor = AsOp_result(profile_photometry_file=a_als, subtracted_image_file=a_sub)
        self._insert_processing_step(commands, output_processor=processor)
        self.ALlstars_result = processor
        if not self.batch_mode:
            self.run()
        return processor



# coding=utf-8
from __future__ import absolute_import, division, print_function
__metaclass__ = type

import os
import sys
import shutil
import hashlib
from copy import deepcopy
try:
    # noinspection PyCompatibility
    from StringIO import StringIO  # python2
except ImportError:
    from io import StringIO  # python3
import subprocess as sp
try:
    # noinspection PyUnresolvedReferences
    from subprocess import TimeoutExpired  # python 3 only
except ImportError:
    pass

#from . import logger as module_logger
from astwro.config.logger import logger as module_logger
from astwro.config import get_config
from astwro.utils import tmpdir, TmpDir
from .output_processors import StreamKeeper, OutputProvider


class Runner(object):
    """
    Base class for specific runners.

    Maintains underlying process lifetime,
    communication with process (streams, output processors chain), runner dir etc...
    """

    class RunnerException(Exception):
        """Exceptions raised by `Runner` and subclasses"""

        def __init__(self, message, runner):
            self.commandline = str(runner.executable) + ' ' + ' '.join(runner.arguments)
            self.stdin = runner.input
            self.stdout = runner.output
            self.stderr = runner.stderr
            super(Runner.RunnerException, self).__init__(message)

        def __str__(self):
            return (super(Runner.RunnerException,self).__str__()
                    + '\n>> Commandline:\n{}\n>> Stdin:\n{}\n>> Stdout:\n{}\n>> Stderr:\n{}\n'.format(
                        self.commandline, self.stdin, self.stdout, self.stderr
                    ))

    class ExitError(RunnerException):
        """Exceptions raised when underlying process returns error code on exit"""
        def __init__(self, message, runner, code):
            super(Runner.ExitError, self).__init__(message, runner)
            self.code = code
        def __str__(self):
            return (super(Runner.ExitError,self).__str__()
                    + '\n>> Process exit code: {}'.format(self.code))

    class NoFileError(RunnerException):
        def __init__(self, message, runner, filename):
            super(Runner.NoFileError, self).__init__(message, runner)
            self.filename = filename

    class RunnerValueError(ValueError, RunnerException):
        pass

    class RunnerTypeError(TypeError, RunnerException):
        pass

    raise_on_nonzero_exitcode = True
    preserve_process = False  # not implemented

    def __init__(self, dir=None, batch=False):
        """
        :param dir: path name or TmpDir object, in not provided new temp dir will be used
        :param bool batch:      whether Daophot have to work in batch mode.
        """
        self.logger = module_logger.getChild(type(self).__name__)
        self.executable = None
        self.arguments = []
        self.batch_mode = batch
        self.__stream_keeper = None

        self._prepare_dir(dir)
        self._reset()

    def _reset(self):
        """Resets runner without cleaning/changing runner dir
           allows execution of new sequence in same dir and files"""
        self.input = None
        self.output = None
        self.stderr = None
        self.returncode = None
        self.__process = None
        self.__commands = ''
        self.ext_output_files = set()

        if self.__stream_keeper is not None:
            self.__stream_keeper.stream = None # new chain containing only old StreamKeeper
        else:
            self.__stream_keeper = StreamKeeper(runner=self)   # new chain containing new StreamKeeper
        self.__processors_chain_last = self.__stream_keeper
        self.__processors_chain_first = None

    def __deepcopy__(self, memo):
        cls = self.__class__
        new = cls.__new__(cls)
        memo[id(self)] = new

        new.__stream_keeper = None
        new._reset()
        new.logger = self.logger
        new.executable = self.executable
        # new.output = self.output
        # new.stderr = self.stderr
        # new.returncode = self.returncode
        new.batch_mode = self.batch_mode
        new.arguments = self.arguments
        # new.__process = None
        # new.__commands = self.__commands
        # new.ext_output_files = set()

        # new.__processors_chain_last  = deepcopy(self.__processors_chain_last, memo) # copy chain
        # new.__processors_chain_first = memo[id(self.__processors_chain_first)]  # find StreamKeeper in copied chain
        # new.__stream_keeper          = memo[id(self.__stream_keeper)]  # find StreamKeeper in copied chain

        new.dir = deepcopy(self.dir, memo)
        return new

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def clone(self):
        """Clones runner

        If *runner directory* was provided in constructor, clone will share the same dir, else, if
        *runner directory* is temp dir created implicitly by runner, clone will create it's own one, and
        content of *runner directory* will be copied from source to clone."""
        return deepcopy(self)

    def close(self):
        """Cleans things up."""
        self._on_exit()
        self.dir = None

    @property
    def mode(self):
        """Either "normal" or "batch". In batch mode, commands are not executed but collected
        on execution queue, then run together, in single process, one by one, 
        triggered by :py:meth:`run()` method"""
        return 'batch' if self.batch_mode else  'normal'

    @mode.setter
    def mode(self, m):
        if m.lowercase() == 'batch':
            self.batch_mode = True
        elif m.lowercase() == 'normal':
            self.batch_mode = False
        else:
            raise Runner.RunnerValueError('mode have to be either "normal" or "batch"', self)

    def _init_workdir_files(self, dir):
        pass

    def _update_executable(self, exe):
        """Find exe key in configuration and set as Runner executable"""
        if self.executable is None:
            self.executable = os.path.expanduser(get_config().get('executables', exe))


    def _prepare_dir(self, dir=None, init_files=True):
        if dir is None:
            dir = tmpdir(prefix='pydaophot_tmp')
        elif isinstance(dir, str):
            dir = tmpdir(use_existing=dir)
        elif not isinstance(dir, TmpDir):
            raise Runner.RunnerTypeError('dir must be either: TmpDir object, str, None', self)
        self.dir = dir
        if init_files:
            self._init_workdir_files(dir)

    def copy_to_runner_dir(self, source, filename=None):
        """Copies source file to  runner dir under name filename or the same
        as original if filename is None. Overwrites existing file."""
        dst = self.dir.path if filename is None else os.path.join(self.dir.path, filename)
        shutil.copy(source, dst)

    def link_to_runner_dir(self, source, link_filename=None):
        # type: (str, str) -> None
        """Creates symlink in  runner dir under name filename or the same
        as original if filename is None. Overwrites existing link.
        :param source: file patch  
        :param link_filename: worker dir link name, default: same as filename part of source"""
        source = self.expand_path(source)
        # if not os.path.isfile(source):
        #     raise IOError('Source file {} not found'.format(source))
        if link_filename is None:
            link_filename = os.path.basename(source)
        dest = os.path.join(self.dir.path, link_filename)
        try:
            os.remove(dest)
        except OSError:
            pass
        os.symlink(source, dest)

    def copy_from_runner_dir(self, filename, dest='./'):
        """Copies file: filename from runner dir. Overwrites existing file."""
        shutil.copy(os.path.join(self.dir.path, filename), dest)

    def link_from_runner_dir(self, filename, dest='./'):
        """Creates symlink in dest of file from runner dir.
        dest can be either file path for new symlink or directory.
        In second case name of symlink will be filename. Overwrites existing file."""
        if os.path.basename(dest) == '':
            dest = os.path.join(dest, filename)
        try:
            os.remove(dest)
        except OSError:
            pass
        os.symlink(os.path.join(self.dir.path, filename), dest)

    def file_from_runner_dir(self, filename):
        """Simply adds runner dir path into filename"""
        return os.path.join(self.dir.path, filename)

    def exists_in_runner_dir(self, filename):
        """Checks for filename existence in runner dir"""
        return os.path.exists(os.path.join(self.dir.path, filename))

    def rm_from_runner_dir(self, filename):
        """Removes (if exists) file filename from runner dir"""
        try:
            os.remove(os.path.join(self.dir.path, filename))
        except OSError:
            pass

    @staticmethod
    def expand_path(path):
        """Expand user ~ directory and finds absolute path."""
        if path is None:
            path = ''
        else:
            path = os.path.abspath(os.path.expanduser(path))
        return path

    def absolute_path(self, path):
        """Returns absolute path for filepath parameter, if :arg:path contain filename only, runner dir is added"""
        if os.path.basename(path) != path:  # not in runner directory
            absolute = self.expand_path(path)
        else:
            absolute = os.path.join(self.dir.path, path)
        return absolute


    @staticmethod
    def _runner_dir_file_name(filepath='', prefix='', suffix='', signature=None, maxlen=30):
        # type: (str, str, str, str, int) -> str
        """Generates name used in Runner local dir for filepath

        Files processed by underlying process are always in it's working directory
          (runner directory). For files from other location in filesystem, copies
          or links in runner directory are maintained. Names of that files are prefixed
          with hash (shortened) of filepath to avoid collisions.
        """
        if signature is None:
            signature = filepath
        basename, ext = os.path.splitext(os.path.basename(filepath))
        maxbasename = maxlen - 7 - len(suffix) - len(prefix) - len(ext)
        if maxbasename < 0:
            raise ValueError('Can not fit working dir filename in manxlen={} characters with prefix={} and suffix={}'
                             .format(maxlen, prefix, suffix))

        basename = basename[: maxbasename]  # cut basename to fit

        return prefix + str(hashlib.md5(str(signature).encode()).hexdigest())[:6] + '_' + basename + suffix + ext

    def _prepare_output_file(self, data, preservefilename=False):
        # type: (str, bool) -> (str, str)
        return self._prepare_io_file(data, output=True, preservefilename=preservefilename)

    def _prepare_input_file(self, path, preservefilename=False):
        # type: (str, bool) -> (str, str)
        return self._prepare_io_file(path, output=False, preservefilename=preservefilename)

    def _prepare_io_file(self, path, output, preservefilename=False):
        # type: (str, bool) -> (str, str)
        """ make link for non-local input files in runner dir, gen runner dir filename """
        if not path:
            return '',''
        if os.path.dirname(os.path.abspath(path)) == self.dir.path: # path to runner dir provided, cut it
            path = os.path.basename(path)
        if os.path.basename(path) != path:  # not in runner directory
            absolute = self.expand_path(path)
            local = os.path.basename(path) if preservefilename else self._runner_dir_file_name(absolute)
            if output:
                # add to list of files to update after run
                self.ext_output_files.add(absolute)
            elif absolute not in self.ext_output_files:
                self.link_to_runner_dir(absolute, local)
                self.logger.debug("Linking input file into runner directory: {} <- {}".format(local, absolute))
        else:
            absolute = os.path.join(self.dir.path, path)
            local = path
        if output:
            # remove runner dir file if exist
            self.rm_from_runner_dir(local)

        return local, absolute

    def _pre_run(self, wait):
        pass

    def run(self, wait=True):
        """
        Execute commands queue.

        In the "normal" :meth:`mode <mode>` there is no need to call :meth:`run`, because all commands are
        executed immediately. In "batch" :meth:`mode <mode>`, commands  execution is queued and postponed
        until :meth:`.run`

        :param bool wait:
            If false,  :meth:`run` exits without waiting for finishing commands executions (asynchronous processing).
            Call :meth:`wait_for_results` before accessing results.
        :return: None
        """
        self._pre_run(wait)
        try:
            self.__process = sp.Popen([self.executable] + self.arguments,
                                      stdin=sp.PIPE,
                                      stdout=sp.PIPE,
                                      stderr=sp.PIPE,
                                      cwd=self.dir.path)
        except OSError as e:
            self.logger.error(
                'Executable: %s is expected in PATH, configure executable name/path in ~/pydaophot.cfg e.g.',
                self.executable)
            raise e
        self.logger.debug('STDIN:\n' + self.__commands)
        if wait:
            self.__communicate(self.__commands)
        else:
            if sys.version_info[0] > 2:  # python 3 has timeout in communicate
                try:
                    self.__communicate(self.__commands, timeout=0.01)
                except TimeoutExpired:
                    pass
            else:  # python 2 - write directly to stdin and close it to flush end generate EOF
                self.__process.stdin.write(self.__commands)
                self.__process.stdin.close()
                self.__process.stdin = None

    def is_ready_to_run(self):
        """
        Returns True if there are some commands waiting for run but process was not started yet

        :return: bool
        """
        return self.__commands and self.__process is None

    @property
    def running(self):
        """
        Whether if runner is running

        ``True`` If executable was started in async mode :meth:`run(wait=False) <run>`, and no output collected yet.

        .. Note::
            Even if executable has finished, output will not be collected and :meth:`running <running>` will
            return ``True`` until user asks for results or call :meth:`wait_for_results()`

        :return: bool
        """
        return self.__process is not None and self.output is None

    def has_finished_run(self):
        """
        Returns True if process has finished and output is available

        :return: bool
        """
        return self.output is not None

    def wait_for_results(self):
        """In the "batch" mode, waits for commands completion if :meth:`run(wait=False) <run>` was called """
        if self.running:
            self.__communicate()
        if self.is_ready_to_run():
            self.run(wait=True)

    def __communicate(self, inpt=None, timeout=None):
        i = inpt.encode(encoding='ascii') if inpt else None
        self.input = i
        o, e = self.__process.communicate(i, timeout=timeout) if timeout else self.__process.communicate(i)
        self.output = o.decode('ascii')
        self.stderr = e.decode('ascii')
        self.logger.debug('STDOUT:\n' + self.output)
        self.__stream_keeper.stream = StringIO(self.output)
        self.returncode = self.__process.returncode
        if self.returncode != 0:
            self.logger.warning('{} process finished with error code {}'.format(self.executable, self.returncode))
            if self.raise_on_nonzero_exitcode:
                raise Runner.ExitError('Execution failed, exit code {}'.format(self.returncode), self, self.returncode)
        # copy results - output files from runners directory to user specified path
        for f in self.ext_output_files:
            try:
                self.copy_from_runner_dir(self._runner_dir_file_name(f), f)
            except FileNotFoundError as e:
                msg = '{} process does not produce expected output file {}<--{}'.format(self.executable, f,
                                                                                        self._runner_dir_file_name(f))
                self.logger.error(msg)
                if self.raise_on_nonzero_exitcode:
                    raise Runner.NoFileError(msg, self, self.returncode)
        # fill chained processors buffers
        self.__processors_chain_last.get_output_stream()


    def _get_ready_for_commands(self):
        if self.running:
            self.wait_for_results()  # if running wait
        if self.has_finished_run():  # if was running reset and get ready for new process
            self._reset()

    def _insert_processing_step(self, std_in, output_processor=None, on_beginning=False):
        if on_beginning:
            self.__commands = std_in + self.__commands
        else:
            self.__commands += std_in
        if output_processor is not None:
            if not isinstance(output_processor, OutputProvider):
                raise Runner.RunnerTypeError('output_processor must OutputProvider subclass', self)
            output_processor.logger = self.logger
            #  chain organisation:
            # [stream_keeper]<-[processors_chain_first]<-[]<-[]<-[processors_chain_last]
            if on_beginning:
                output_processor._prev_in_chain = self.__stream_keeper
                if self.__processors_chain_first is not None:
                    self.__processors_chain_first._prev_in_chain = output_processor
                self.__processors_chain_first = output_processor
            else:
                output_processor._prev_in_chain = self.__processors_chain_last
                self.__processors_chain_last = output_processor
                if self.__processors_chain_first is None:
                    self.__processors_chain_first = output_processor
        return output_processor

    def _on_exit(self):
        pass



# coding=utf-8
from __future__ import absolute_import, division, print_function
__metaclass__ = type

import shutil
from tempfile import mkdtemp
from copy import deepcopy
from os import PathLike

class TmpDir(PathLike):
    """
    instances of TmpDir keeps track and lifetime of temporary directory
    """
    path = None
    dir_is_tmp = True
    _prefix = ''
    _base = None

    def __init__(self, use_existing=None, prefix='astwro_tmp_', base_dir=None):
        """
        :param str use_existing:    If provided, instance will point to that directory and not delete it on destruct
        :param str prefix:          Prefix for temporary dir
        :param str base_dir:        Where to crate tem dir, in None system default is used
        """
        self._prefix = prefix
        self._base = base_dir
        if use_existing is None:
            self.path = mkdtemp(prefix=prefix, dir=base_dir)
            self.dir_is_tmp = True
        else:
            self.path = use_existing
            self.dir_is_tmp = False

    def clone(self):
        return deepcopy(self)

    def __del__(self):
        self._rm_dir()

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        self._rm_dir()

    def __repr__(self):
        return ('Tmp dir:' if self.dir_is_tmp else 'Ext dir:') + (self.path if self.path else 'none')

    def __str__(self):
        return self.path

    def __deepcopy__(self, memo):
        cls = self.__class__
        new = cls.__new__(cls)
        memo[id(self)] = new
        if self.dir_is_tmp:
            new.__init__(prefix=self._prefix, base_dir=self._base)
            shutil.rmtree(new.path)
            shutil.copytree(self.path, new.path, symlinks=True)
        else:
            new.__init__(use_existing=self.path)
        return new

    def __fspath__(self):
        """Implements ``os.PathLike interface``"""
        return str(self)

    def _rm_dir(self):
        """Deletes working dir with all content."""
        if self.dir_is_tmp:
            try:
                shutil.rmtree(self.path)
            except OSError:
                pass





class Individual(object):

    genome = None
    __score = None

    def __init__(self, score_function=None):
        """
        :param ScoreFunction score_function:
        """
        self.score_function = score_function

    @property
    def score(self):
        if self.__score is None:
            self.__score = self.score_function.calc(individual=self)
        return self.__score



# coding=utf-8
from __future__ import absolute_import, division, print_function

__metaclass__ = type
# see https://wiki.python.org/moin/PortingToPy3k/BilingualQuickRef

from os.path import join
import numpy as np
from astropy.table import Table
from astropy.modeling import models, fitting
from astropy.stats import sigma_clipped_stats, sigma_clip

from astwro.exttools import Runner
from astwro.phot.lc_io import write_lc, write_lc_filters

class LCFitter(object):
    """
    Abstract Light Curve Fitter

    Base class for light curve fitters used by `FNpeaksResult.makefit`

    Parameters
    ----------
    clipping_sigma: float
        Sigma clipping to this of light curve before fitting
    """
    def __init__(self, clipping_sigma=None):
        self.clipping_sigma = clipping_sigma
        self.fit = None


    def __call__(self, *args, **kwargs):
        """
        Call underlying model
        """
        return self.fit(*args, **kwargs)

    def dofit(self, freq, hjd, lc, lc_e=None):
        """
        Performs model fit to light curve

        Dummy abstract method to be overridden

        Parameters
        ----------
        freq: float
            Frequency found by FNpeaks, fixed model parameter.
        hjd : array-like
            N-element time axis
        lc : array-like
            N-element measurements on points on `hjd` time moments
        lc_e : array-like
            N-element optional errors of measurements `lc`

        Returns
        -------
        Touple clipped_hjd, fitted_model
        """
        raise NotImplementedError('Abstract LCFitter.fit method called')
        return None, self


class SineFitter(LCFitter):
    """A sin(2pi*f + phi)"""
    def __init__(self, clipping_sigma=None):
        super(SineFitter, self).__init__(clipping_sigma)
        self.fit = None
        self.frequency = None
        self.amplitude = None
        self.phase = None


    def dofit(self, freq, hjd, lc, lc_e=None):
        fitter = fitting.FittingWithOutlierRemoval(fitting.LevMarLSQFitter(), sigma_clip, sigma=self.clipping_sigma)
        model = models.Sine1D(frequency=freq)
        model.frequency.fixed = True
        clipped, fit = fitter(model, hjd, lc - sigma_clipped_stats(lc)[0])
        self.fit = fit
        self.frequency = fit.frequency
        self.amplitude = fit.amplitude
        self.phase = fit.phase
        return clipped, self

class SineFitterC(LCFitter):
    """A sin(2pi*f + phi) + mag"""
    def __init__(self, clipping_sigma=None):
        super(SineFitterC, self).__init__(clipping_sigma)
        self.fit = None
        self.frequency = None
        self.amplitude = None
        self.phase = None
        self.mag = None

    def dofit(self, freq, hjd, lc, lc_e=None):
        fitter = fitting.FittingWithOutlierRemoval(fitting.LevMarLSQFitter(), sigma_clip, sigma=self.clipping_sigma)
        model = models.Sine1D(frequency=freq) + models.Const1D(0)
        model.frequency_0.fixed = True
        clipped, fit = fitter(model, hjd, lc)
        self.fit = fit
        self.frequency = fit.frequency_0
        self.amplitude = fit.amplitude_0
        self.phase = fit.phase_0
        self.mag = fit.amplitude_1
        return clipped, self


class SineHarmonicFitterC(LCFitter):
    """A SUM_i sin(2pi*f*i + phi) + mag"""
    def __init__(self, clipping_sigma=None, minn=1, maxn=6):
        super(SineHarmonicFitterC, self).__init__(clipping_sigma)
        self.fit = None
        self.frequency = None
        self.amplitude = None
        self.phase = None
        self.mag = None
        self.minn = minn
        self.maxn = maxn

    def dofit(self, freq, hjd, lc, lc_e=None):
        inner_fitter = fitting.LevMarLSQFitter()
        fitter = fitting.FittingWithOutlierRemoval(inner_fitter, sigma_clip, sigma=self.clipping_sigma)
        ret_clipped = None
        for n in range(self.minn, self.maxn+1):
            model = models.Const1D(0)
            for i in range(1, n+1): # add i harmonics to model
                model_s = models.Sine1D(frequency=freq*i, amplitude=0)
                model_s.frequency.fixed = True
                model = model + model_s
            clipped, fit = fitter(model, hjd, lc)
            if ret_clipped is None:
                ret_clipped = clipped  # return first clipping not harmonics clipping
            chi = np.ma.std(clipped - fit(hjd))
            print(chi)
        self.fit = fit
        self.frequency = fit.frequency_1
        self.amplitude = fit.amplitude_1
        self.phase = fit.phase_1
        self.mag = fit.amplitude_0
        return ret_clipped, self

class Peak(object):
    """
    Peak description.

    Parameters
    ----------
    freq : float
    period : float
    power : float
    sn : float
        Signal to noise of peak
    """
    def __init__(self, freq, period, power, sn):
        self.freq = freq
        self.period = period
        self.power = power
        self.sn = sn


class FNpeaksResult(object):
    """
    Results of fnpeak call.

    Returned by `~astwro.timeseries.FNpeaks`.

    Attributes
    ----------
    power : `~numpy.ndarray`
        Power of signal in frequencies, use `~astwro.timeseries.FNpeaks.freq` method for i-th frequency
    peaks : list of `~astwro.timeseries.Peak`
        List of most prominent peaks returned by fnpeaks (*.max file)
        notes).

    """
    def __init__(self):
        self.power = None
        self.peaks = None
        self.lc = None
        self.lc_e = None
        self.hjd = None
        self.freq_range = (0.0, 0.0)
        self.freq_step = 0.0
        self.fit = None
        self.lc_clipped = None

    @property
    def frequencies(self):
        return np.arange(self.freq_range[0], self.freq_range[1], self.freq_step)

    @property
    def periods(self):
        return 1.0 / np.arange(self.freq_range[0], self.freq_range[1], self.freq_step)

    @property
    def lc_reduced(self):
        if self.fit is None:
            self.makefit()
        return self.lc_clipped - self.fit(self.hjd)

    def makefit(self, freq=None, peak=0, clipping_sigma=3.0, fitter_class=SineHarmonicFitterC, **kwargs):
        if freq is None:
            freq = self.peaks[peak]['freq']
        fitter = fitter_class(clipping_sigma, **kwargs)
        self.lc_clipped, self.fit = fitter.dofit(freq, self.hjd, self.lc,self.lc_e)
        return self.fit

    # def makefit(self, freq=None, peak=0, clipping_sigma=3.0):
    #     if freq is None:
    #         freq = self.peaks[peak]['freq']
    #     fitter = fitting.FittingWithOutlierRemoval(fitting.LevMarLSQFitter(), sigma_clip, sigma=clipping_sigma)
    #     model = models.Sine1D(frequency=freq)
    #     model.frequency.fixed = True
    #     self.lc_clipped, self.fit = fitter(model, self.hjd, self.lc - sigma_clipped_stats(self.lc)[0])
    #     return self.fit



class FNpeaks(Runner):
    """
    `fnpeaks` runner

    Object of this class maintains single process of `fnpeaks` and it's working directory.

    Parameters
    ----------
    hjd : `~numpy.ndarray`
        Heliocentric julian days for observations (second dimension of lc)
    lc : `~numpy.ndarray`
        2D NxM array-like with light curves. N stars, M observations
    lc_e : `~numpy.ndarray`, optional
        Standart deviations of `lc`
    filtermasks : dict, optional
        Keys are filters, values are boolean masks for second dimension of `lc` with masks of filter
    start_freq : float
        Start frequency to scan (1/day)
    end_freq : float
        End frequency to scan (1/day)
    step_freq : float
        Frequency step
    ids : array-like:
        1D N element array with numerical ids of stars for filenames generation
    dir : str, optional
        Path of directory used for fnpeaks input output. If not provided temporary dir will be used
        and deleted on destructor.
    """

    def __init__(self, hjd, lc, lc_e=None, filtermasks=None,
                 start_freq=0.0001, end_freq=40.0001, step_freq=0.0001,
                 ids=None, dir=None):
        super(FNpeaks, self).__init__(dir=dir)
        self.hjd = hjd
        if lc.ndim == 1:
            self.lc = lc.reshape((1, lc.size))
        else:
            self.lc = lc
        if lc_e is not None and lc_e.ndim == 1:
            self.lc_e = lc_e.reshape((1, lc_e.size))
        else:
            self.lc_e = lc_e
        self.filtermasks = filtermasks
        self.start_freq = start_freq
        self.end_freq = end_freq
        self.step_freq = step_freq
        self.start_freq = start_freq
        self.ids = ids
        self._update_executable('fnpeaks')

    def __call__(self, star, filter=None, periodogram=False, error_max=None, sigmaclip=None):
        """
        Runs fnpeaks for specified star(s)

        Parameters
        ----------
        star : long or array-like
            Stars(s) to process, position index in `~astwro.timeseries.FNpeaks.lc`
        filter : list of str
           If `~astwro.timeseries.FNpeaks.filtermasks` specified, selects filters. None means: all filters.
        periodogram : bool
            Whether to generate power spectral density estimates
        """
        # TODO reuse saved files
        if isinstance(star, int):
            star = [star]
        lc = self.lc[star]
        lc_e = None
        if self.lc_e is not None:
            lc_e = self.lc_e[star]
        ids = None
        if self.ids is not None:
            ids = self.ids[star]
        else:
            ids = star

        mask = np.ones_like(lc, dtype=bool)
        if error_max is not None and lc_e is not None:
            mask = lc_e < error_max
            lc = np.ma.MaskedArray(lc)
            lc.mask = lc.mask | ~(lc_e < error_max)
        if sigmaclip is not None:
            lc = sigma_clip(lc, sigma=sigmaclip)

        if self.filtermasks is None:
            write_lc(self.hjd, lc, lc_e, ids, 0.0,
                     prefix=join(self.dir.path, 'lc_'), suffix='.diff')
        else:
            if filter is None:
                filter = [f for f in self.filtermasks]
            write_lc_filters(self.filtermasks, filter, self.hjd, lc, lc_e, ids, 0.0,
                             prefix=join(self.dir.path, 'lc_'), suffix='.diff')
        ret = []
        for s in star:
            r = None
            if self.filtermasks is None:
                r = self._runfnpeaks(s, None, periodogram)
            else:
                r = {}
                for f in filter:
                    r[f] = self._runfnpeaks(s, f, periodogram)
            ret.append(r)
        if len(ret) == 1:
            ret = ret[0]
        return ret

    def freq(self, i):
        """
        Returns i-th frequency(ies)
        """
        return self.start_freq + i * self.step_freq

    def period(self, i):
        """
        Returns i-th period(s)
        """
        return 1.0 / self.freq(i)

    def _runfnpeaks(self, star_pos, filter, periodogram):
        if self.ids is not None:
            star = np.array(self.ids)[star_pos]
        else:
            star = star_pos
        if filter is None:
            fname = 'lc_{:05d}'.format(star)
        else:
            fname = 'lc_{:05d}_{}'.format(star, filter)
        if periodogram:
            self.arguments = ['-f']
        else:
            self.arguments = []
        self.arguments += [fname+'.diff']
        self.arguments += ['{:f}'.format(x) for x in [self.start_freq, self.end_freq, self.step_freq]]
        self.run(wait=True)

        ret = FNpeaksResult()
        with open(join(self.dir.path, fname)+'.max', 'r') as f:
            lines = f.readlines()
        peaks = Table(names=['no', 'freq', 'period', 'amplitude', 'sn'])
        for i, l in enumerate(lines[9:]):
            if len(l) > 2:
                peaks.add_row([float(s) for s in l.split()])
        ret.peaks = peaks
        ret.freq_range = (self.start_freq, self.end_freq)
        ret.freq_step = self.step_freq
        ret.lc   = self.lc  [star_pos]  #TODO cache lc[:,filter] in internal dict via @property _fmask_lc(f)
        ret.lc_e = self.lc_e[star_pos] if self.lc_e is not None else None
        ret.hjd  = self.hjd
        if filter is not None and self.filtermasks is not None:
            ret.lc   = ret.lc  [self.filtermasks[filter]]
            ret.lc_e = ret.lc_e[self.filtermasks[filter]] if ret.lc_e is not None else None
            ret.hjd  = ret.hjd [self.filtermasks[filter]]
        if periodogram:
            ret.power = np.fromfile(join(self.dir.path, fname)+'.trf', sep=' ')[1::2] # only freq column
        return ret






#! /usr/bin/env python
# coding=utf-8

from __future__ import print_function, division

from scipy.stats import sigmaclip

from astwro.pydaophot import daophot
from astwro.pydaophot import fname
from astwro.pydaophot import allstar
from astwro.starlist import read_dao_file
from astwro.starlist import write_ds9_regions
import __commons as commons

# TODO: expand this script to (optionally) leave result files - make it allstar runner
# TODO: implement creating ds9 region (why? or remove that option)


def __do(args):
    # 1 do daophot aperture and psf photometry and run allstar

    dp = daophot(image_file=args.image)
    dp.copy_to_runner_dir(args.coo, fname.COO_FILE)
    dp.PHotometry()
    dp.copy_to_runner_dir(args.lst, fname.LST_FILE)
    dp.PSf()
    dp.run(wait=True)
    al = allstar(dp.dir)
    al.run()
    all_s = read_dao_file(al.file_from_runner_dir(fname.ALS_FILE))
    # all_s.hist('chi')
    return sigmaclip(all_s.chi)[0].mean()

    # 2 write regions
    # if args.regions:
    #
    #     write_ds9_regions(coo_file+'.reg', )


def __arg_parser():
    import argparse
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=commons.version_string(),
        description='Runs daophot photometry and allstar. Find mean chi of allstar stars. \n\n'
                    'Takes FITS image and star lists: coo - all stars, lst - PSF stars\n'
                    'then performs daophot\'s PH, PS and allstar on them.\n'
                    '(chi is calculated by allstar for every star as \n'
                    '  \"the observed pixel-to-pixel scatter from the model image profile \n'
                    '  DIVIDED BY the expected pixel-to-pixel scatter from the image profile\").\n'
                    'The same mean chi is used as function to be minimized by genetic \n'
                    'algorithm in gapick.py. This script allows quick comparison \n'
                    'between different PSF stars sets.',)
    parser.add_argument('image', type=str,
                        help='FITS image file')
    parser.add_argument('coo', type=str,
                        help='all stars list: coo file')
    parser.add_argument('lst', type=str,
                        help='PSF stars list: lst file')
    parser.add_argument('--reg', action='store_true',
                        help='create ds9 region files <name>.coo.reg and <name>.lst.reg')
    return parser

# Below: standard skeleton for astwro.tools


def main(image, coo, lst, **kwargs):
    args = commons.bunch_kwargs(__arg_parser(), positional=[image, coo, lst], **kwargs)
    return __do(args)


def info():
    commons.info(__arg_parser())


if __name__ == '__main__':

    __args = __arg_parser().parse_args()
    print(__do(__args))
